"""
Tools for retrieving price data.
Currently supports Coingecko and locally stored data.

Note
-----
Nomics data is deprecated.
"""

from curvesim.exceptions import NetworkError

from .sources import coingecko, local


def get(
    coins,
    chain="mainnet",
    *,
    days=60,
    data_dir="data",
    src="coingecko",
    end=None,
):
    """
    Pull price and volume data for given coins.

    Data is returned for all pairwise combinations of the input coins.

    Parameters
    ----------
    coins : list of str
        List of coin addresses.

    days : int, default=60
        Number of days to pull data for.

    data_dir : str, default="data"
        Directory to load local data from.

    src : str, default="coingecko"
        Data source ("coingecko", "nomics", or "local").


    Returns
    -------
    prices : pandas.DataFrame
        Timestamped prices for each pair of coins.

    volumes : pandas.DataFrame
        Timestamped volumes for each pair of coins.

    pzero : int or pandas.Series
        Proportion of timestamps with zero volume.

    """
    if src == "coingecko":
        prices, volumes, pzero = coingecko(coins, chain=chain, days=days, end=end)

    elif src == "nomics":
        raise NetworkError("Nomics data is no longer supported.")

    elif src == "local":
        prices, volumes, pzero = local(coins, data_dir=data_dir, end=end)

    return prices, volumes, pzero



"""Miscellaneous utility functions."""
from math import factorial

from numpy import append, array

from curvesim.logging import get_logger
from curvesim.pool import CurveCryptoPool, CurveMetaPool, CurvePool

logger = get_logger(__name__)


def compute_volume_multipliers(pool_vol, market_vol, n, pool_type, mode=1):
    """
    Computes volume multipliers (vol_mult) used for volume limiting.

    Parameters
    ----------
    pool_vol : float, int, or list
        Total volume for the pool over the simulation period.

    market_vol : dict
        Total market volume for each token pair over the simulation period.

    n : int
        The number of token-types in the pool (e.g., DAI, USDC, USDT = 3)

    pool_type : str
        "CurvePool" or "CurveMetaPool"

    vol_mode : int, default=1
        Modes for computing the volume multiplier:

        1: limits trade volumes proportionally to market volume for each pair
        2: limits trade volumes equally across pairs
        3: mode 2 for trades with meta-pool asset, mode 1 for basepool-only trades

    """
    pairs, market_vol = zip(*market_vol.items())
    pool_vol_array = array(pool_vol)
    market_vol_array = array(market_vol)

    try:
        get_vol_mult = _pool_functions[pool_type]
    except KeyError as e:
        raise TypeError(
            f"Pool type {pool_type} not supported by volume limiter."
        ) from e

    vol_mult = get_vol_mult(pool_vol_array, market_vol_array, n, mode)
    vol_mult_dict = dict(zip(pairs, vol_mult))

    logger.info("Volume Multipliers: %s", _format_info_str(vol_mult_dict))
    return vol_mult_dict


def _pool_vol_mult(pool_vol, market_vol, n, mode):
    if mode == 1:
        vol_mult = [pool_vol / market_vol.sum()] * n

    elif mode == 2:
        vol_mult = pool_vol.repeat(n) / n / market_vol

    elif mode == 3:
        logger.info("Vol_mode=3 only available for meta-pools. Reverting to vol_mode=1")
        vol_mult = [pool_vol / market_vol.sum()] * n

    else:
        raise ValueError(f"Mode must be integer 1, 2, or 3, not {mode}.")

    return vol_mult


def _metapool_vol_mult(pool_vol, market_vol, n, mode):
    pool_vol_meta = pool_vol[0]
    pool_vol_base = pool_vol[1]
    mkt_vol_meta = market_vol[0 : n[1]]
    mkt_vol_base = market_vol[n[1] :]

    n_base_pairs = int(factorial(n[1]) / 2 * factorial(n[1] - 2))

    if mode == 1:
        vol_mult = append(
            pool_vol_meta / mkt_vol_meta.sum().repeat(n[1]),
            pool_vol_base / mkt_vol_base.sum().repeat(n_base_pairs),
        )

    elif mode == 2:
        vol_mult = append(
            pool_vol_meta.repeat(n[1]) / n[1] / mkt_vol_meta,
            pool_vol_base.repeat(n_base_pairs) / n_base_pairs / mkt_vol_base,
        )

    elif mode == 3:
        vol_mult = append(
            pool_vol_meta.repeat(n[1]) / n[1] / mkt_vol_meta,
            pool_vol_base / mkt_vol_base.sum().repeat(n_base_pairs),
        )

    else:
        raise ValueError(f"Mode must be integer 1, 2, or 3, not {mode}.")

    return vol_mult


def _format_info_str(vol_mult_dict):
    info = [f"{base}/{quote}: {mult}" for (base, quote), mult in vol_mult_dict.items()]
    new_line = "\n    "
    return new_line + new_line.join(info)


_pool_functions = {
    CurvePool: _pool_vol_mult,
    CurveMetaPool: _metapool_vol_mult,
    CurveCryptoPool: _pool_vol_mult,
}



"""
A simulation runs trades against Curve pools, using a strategy that may
utilize different types of informed or noise trades.

The :mod:`simulation pipeline framework <curvesim.pipelines>` allows the
user to build custom strategies for simulation.

Most users will want to use the `autosim` function, which supports
"optimal" arbitrages via the
:func:`volume-limited arbitrage pipeline <curvesim.pipelines.vol_limited_arb.pipeline>`.
The primary use-case is to determine optimal amplitude (A) and fee
parameters given historical price and volume feeds.
"""
from curvesim.logging import get_logger
from curvesim.pipelines.vol_limited_arb import DEFAULT_PARAMS
from curvesim.pipelines.vol_limited_arb import pipeline as volume_limited_arbitrage
from curvesim.pool_data import get_metadata
from curvesim.utils import get_pairs

logger = get_logger(__name__)


def autosim(
    pool=None,
    chain="mainnet",
    pool_metadata=None,
    pool_data_cache=None,
    env="prod",
    **kwargs,
):
    """
    The autosim() function simulates existing Curve pools with a range of
    parameters (e.g., the amplitude parameter, A, and/or the exchange fee).

    The function fetches pool properties (e.g., current pool size) and 2
    months of price/volume data and runs multiple simulations in parallel.

    Curve pools from any chain supported by the Convex Community Subgraphs
    can be simulated directly by inputting the pool's address.

    Parameters
    ----------
    pool: str, optional
        This 0x-prefixed string identifies the pool by address.

        .. note::
            Either `pool` or `pool_metadata` must be provided.

    chain: str, default='mainnet'
        Identifier for blockchain or layer2.  Supported values are:
            "mainnet", "arbitrum", "optimism", "fantom", "avalanche"
            "matic", "xdai"

    pool_metadata: PoolMetaDataInterface, optional
        Pool state and metadata necessary to instantiate a pool object.

        .. note::
            Either `pool` or `pool_metadata` must be provided.

    pool_data_cache: PoolDataCache, optional
        Cached data used in sims.  Useful for replication of results and
        avoiding re-fetches of data.

    A: int or iterable of int, optional
        Amplification coefficient.  This controls the curvature of the
        stableswap bonding curve.  Increased values makes the curve
        flatter in a greater neighborhood of equal balances.

        Defaults to ``[int(2 ** (a / 2)) for a in range(12, 28)]``.

        For basepool, use **A_base**.

    D: int, optional
        Total pool liquidity given in 18 decimal precision.
        Defaults to on-chain data.

        For basepool, use **D_base**.

    tokens: int, optional
        Total LP token supply.
        Defaults to on-chain data.

        For basepool, use **tokens_base**.

    fee: int or iterable of int, optional
        Fees taken for both liquidity providers and the DAO.

        Units are in fixed-point so that 10**10 is 100%,
        e.g. 4 * 10**6 is 4 bps and 2 * 10**8 is 2%.

        Defaults to ``list(range(1000000, 5000000, 1000000))``.

        For basepool, use **fee_base**.

    fee_mul : int
        Fee multiplier for dynamic fee pools.

        For basepool, use **fee_mul_base**.

    admin_fee : int, default=0 * 10**9
        Fees taken for the DAO.  For factory pools, it is half of the total
        fees, as was typical for previous non-factory pools.

        Units are fixed-point percentage of `fee`, e.g. 5 * 10**9 is
        50% of the total fees.

    test: bool, default=False
        Overrides variable_params to use four test values:

        .. code-block::

            {"A": [100, 1000], "fee": [3000000, 4000000]}

    days: int, default=60
        Number of days to fetch data for.

    src: str, default='coingecko'
        Valid values for data source are 'coingecko' or 'local'

    data_dir: str, default='data'
        Relative path to saved data folder.

    vol_mult : dict, float, or int, default computed from data
        Value(s) multiplied by market volume to specify volume limits
        (overrides vol_mode).

        dict should map from trade-pair tuples to values, e.g.:

        .. code-block::

            {('DAI', 'USDC'): 0.1, ('DAI', 'USDT'): 0.1, ('USDC', 'USDT'): 0.1}

    vol_mode : int, default=1
        Modes for limiting trade volume.
        1: limits trade volumes proportionally to market volume for each pair
        2: limits trade volumes equally across pairs
        3: mode 2 for trades with meta-pool asset, mode 1 for basepool-only trades

    ncpu : int, default=os.cpu_count()
        Number of cores to use.

    env: str, default='prod'
        Environment for the Curve subgraph, which pulls pool and volume snapshots.

    Returns
    -------
    dict
        Dictionary of results, each value being a pandas.Series.
    """
    assert any([pool, pool_metadata]), "Must input 'pool' or 'pool_metadata'"

    pool_metadata = pool_metadata or get_metadata(pool, chain, env)
    p_var, p_fixed, kwargs = _parse_arguments(pool_metadata, **kwargs)

    results = volume_limited_arbitrage(
        pool_metadata,
        pool_data_cache,
        variable_params=p_var,
        fixed_params=p_fixed,
        **kwargs,
    )

    return results


def _parse_arguments(pool_metadata, **kwargs):
    pool_args = [
        "A",
        "D",
        "balances",
        "fee",
        "fee_mul",
        "tokens",
        "admin_fee",
        "gamma",
        "fee_gamma",
        "mid_fee",
        "out_fee",
    ]
    pool_args += [arg + "_base" for arg in pool_args[:-1]]

    variable_params = {}
    fixed_params = {}
    rest_of_params = {}
    defaults = DEFAULT_PARAMS.copy()

    for key, val in kwargs.items():
        if key in defaults:
            del defaults[key]

        if key in pool_args:
            if isinstance(val, int):
                fixed_params[key] = val

            elif all(isinstance(v, int) for v in val):
                variable_params[key] = val

            else:
                raise TypeError(f"Argument {key} must be an int or iterable of ints")

        elif key == "vol_mult" and isinstance(val, (int, float)):
            coin_pairs = get_pairs(pool_metadata.coin_names)
            rest_of_params[key] = dict.fromkeys(coin_pairs, val)

        else:
            rest_of_params[key] = val

    variable_params = variable_params or defaults

    return variable_params, fixed_params, rest_of_params



from curvesim.exceptions import CurvesimValueError
from curvesim.logging import get_logger
from curvesim.network.subgraph import redemption_prices_sync as _redemption_prices
from curvesim.network.subgraph import volume_sync as _volume
from curvesim.pool.stableswap.metapool import CurveMetaPool
from curvesim.pool_data.metadata.base import PoolMetaDataInterface

from .metadata import PoolMetaData

logger = get_logger(__name__)


class PoolDataCache:
    """
    Container for fetching and caching historical volume and redemption price data.

    Deprecation warning: this will likely be removed in a future release.
    """

    def __init__(self, metadata_dict, cache_data=False, days=60, end=None):
        """
        Parameters
        ----------
        metadata_dict : dict, :class:`PoolMetaDataInterface`
            Pool metadata in the format returned by
            :func:`curvesim.network.subgraph.pool_snapshot`.

        cache_data : bool, optional
            If True, fetches and caches historical volume and redemption price.

        days : int, default=60
            Number of days to pull data for if caching.

        """
        if isinstance(metadata_dict, dict):
            self.metadata = PoolMetaData(metadata_dict)
        elif isinstance(metadata_dict, PoolMetaDataInterface):
            self.metadata = metadata_dict
        else:
            raise CurvesimValueError(
                "Metadata must be of type dict or PoolMetaDataInterface."
            )

        self.days = days
        self.end = end

        self._cached_volume = None
        self._cached_redemption_prices = None

        if cache_data:
            self.set_cache()

    def set_cache(self):
        """
        Fetches and caches historical volume and redemption price data.

        Parameters
        ----------
        days : int, default=60
            number of days to pull data for
        """
        self._cached_volume = self._get_volume()
        self._cached_redemption_prices = self._get_redemption_prices()

    def clear_cache(self):
        """
        Clears any cached data.
        """
        self._cached_volume = None
        self._cached_redemption_prices = None

    @property
    def volume(self):
        """
        Fetches the pool's historical volume over the specified number of days.

        Parameters
        ----------
        days : int, default=60
            Number of days to pull data for.

        store : bool, default=False
            If true, caches the fetched data.

        get_cache : bool, default=True
            If true, returns cached data when available.

        Returns
        -------
        numpy.ndarray
            Total volume summed across the specified number of days.

        """
        if self._cached_volume is not None:
            logger.info("Getting cached historical volume...")
            return self._cached_volume

        return self._get_volume()

    def _get_volume(self):
        logger.info("Fetching historical volume...")
        addresses = self.metadata.address
        chain = self.metadata.chain
        days = self.days
        end = self.end

        if issubclass(self.metadata.pool_type, CurveMetaPool):
            # pylint: disable-next=protected-access
            basepool_address = self.metadata._dict["basepool"]["address"]
            addresses = [addresses, basepool_address]
            vol = _volume(addresses, chain, days=days, end=end)
            summed_vol = [sum(v) for v in vol]
        else:
            vol = _volume(addresses, chain, days=days, end=end)
            summed_vol = sum(vol)

        return summed_vol

    @property
    def redemption_prices(self):
        """
        Fetches the pool's redemption price over the specified number of days.

        Note: only returns data for RAI3CRV pool. Otherwise, returns None.

        Parameters
        ----------
        days : int, default=60
            Number of days to pull data for.

        store : bool, default=False
            If True, caches the fetched data.

        get_cache : bool, default=True
            If True, returns cached data when available.

        Returns
        -------
        pandas.DataFrame
            Timestamped redemption prices across the specified number of days.

        """
        if self._cached_redemption_prices is not None:
            logger.info("Getting cached redemption prices...")
            return self._cached_redemption_prices

        return self._get_redemption_prices()

    def _get_redemption_prices(self):
        address = self.metadata.address
        chain = self.metadata.chain

        days = self.days
        end = self.end

        r = _redemption_prices(address, chain, days=days, end=end)

        return r



"""
This module provides functionality for creating interactive selector charts with
Altair and pandas.

It contains utility functions for creating selector charts, initializing selectors,
manipulating opacity of selections, and conditionally applying properties.
"""
from altair import Axis, Chart, Color, Scale, condition, selection_point, value
from pandas import DataFrame

from curvesim.exceptions import PlotError

from .chart_properties import PROPERTY_CLASSES


def make_selector(
    field,
    options,
    *,
    labels=None,
    title=None,
    style=None,
    toggle=True,
    sel_idx=None,
):
    """
    Create an interactive selector chart and an Altair selection object.

    Parameters
    ----------
    field : str
        The field in the data to bind the selection to.
    options : list
        The options for the selection.
    labels : list, optional
        The labels for the options. Defaults to the options themselves.
    title : str, optional
        The title of the chart. Defaults to an empty string.
    style : dict, optional
        A dictionary of style options to update the default style with.
    toggle : bool, optional
        Whether to allow toggle behavior in the selection. Defaults to True.
    sel_idx : int or 'all', optional
        The initial selection index. If an integer, select that option. If 'all',
        select all options. If None, no initial selection. Defaults to None.

    Returns
    -------
    altair.Chart
        The created selector chart.
    altair.Selection
        The created selection object.

    Raises
    ------
    PlotError
        If sel_idx is not None, an integer, or 'all'.
    """
    title = title or ""
    labels = labels or options

    init_sel = get_initial_selection(field, options, sel_idx)
    selector = selection_point(
        fields=[field], clear=False, value=init_sel, toggle=toggle
    )

    sel_chart = make_selector_chart(field, options, labels, selector, style)

    return sel_chart.properties(title=title), selector


def get_initial_selection(field, options, sel_idx):
    """
    Get the initial selection value based on provided options and an index.

    Parameters
    ----------
    field : str
        The field in the data to bind the selection to.
    options : list
        The options for the selection.
    sel_idx : int or 'all', optional
        The initial selection index. If an integer, select that option. If 'all',
        select all options. If None, no initial selection.

    Returns
    -------
    dict or list of dict or None
        The initial selection value. If sel_idx is an integer, a dictionary. If 'all',
        a list of dictionaries. If None, None.

    Raises
    ------
    PlotError
        If sel_idx is not None, an integer, or 'all'.
    """
    if sel_idx is None:
        init_sel = None

    elif isinstance(sel_idx, int):
        init_sel = [{field: options[sel_idx]}]

    elif sel_idx == "all":
        init_sel = [{field: o} for o in options]

    else:
        raise PlotError(f"sel_idx must be None or int or 'all', not {type(sel_idx)}.")

    return init_sel


def make_selector_chart(field, options, opt_labels, selector, style=None):
    """
    Create the selector chart.

    Parameters
    ----------
    field : str
        The field in the data to bind the selection to.
    options : list
        The options for the selection.
    opt_labels : list
        The labels for the options.
    selector : altair.Selection
        The selector to add to the chart.
    style : dict, optional
        A dictionary of style options to update the default style with.

    Returns
    -------
    altair.Chart
        The created selector chart.
    """
    _style = {
        "axis": "y",
        "orient": "right",
        "color": Color(scale=Scale(scheme="viridis"), legend=None),
    }

    if style:
        _style.update(style)

    axis = _style["axis"]
    orient = _style["orient"]
    axis_class = PROPERTY_CLASSES["encoding"][axis]

    encoding = {
        axis: axis_class("labels:O", title=None, axis=Axis(orient=orient)),
        "color": _style["color"],
        "opacity": condition(selector, value(1), value(0.1)),
    }

    df = DataFrame({field: options, "labels": opt_labels})
    return Chart(df).mark_rect().encode(**encoding).add_params(selector)


def if_selected(selected, selector, field, if_true, if_false):
    """
    Apply a property if a selection is active.

    Parameters
    ----------
    selected : any
        The selected value.
    selector : altair.Selection
        The selector to check.
    field : str
        The field in the data to check the selection against.
    if_true : any
        The value to return if the selection is active.
    if_false : any
        The value to return if the selection is not active.

    Returns
    -------
    altair.condition
        The condition to apply to a chart.
    """
    return condition(
        f"indexof({selector.name}.{field}, '{selected}') != -1", if_true, if_false
    )



from abc import ABC, abstractmethod
from datetime import datetime

from curvesim.logging import get_logger
from curvesim.utils import dataclass

logger = get_logger(__name__)


@dataclass(slots=True)
class PriceSample:
    """
    Attributes
    -----------
    timestamp : datetime.datetime
        Timestamp for the current price/volume.
    prices : dict
        Price for each pairwise coin combination.
    """

    timestamp: datetime
    prices: dict


class PriceSampler(ABC):
    """
    An iterator that yields market data ticks, i.e. price or other data
    for each timestamp.
    """

    @abstractmethod
    def __iter__(self) -> PriceSample:
        """
        Yields
        -------
        class:`PriceSample`
        """
        raise NotImplementedError



from abc import ABC, abstractmethod
from contextlib import contextmanager

from curvesim.exceptions import SnapshotError


class SnapshotMixin:
    """
    Pool mixin to allow ability to snapshot partial states
    and revert back to it.

    Main class must have `snapshot_class` attribute, which
    implements the `Snapshot` interface.
    """

    snapshot_class = None

    def get_snapshot(self):
        """Saves the pool's partial state."""
        if not self.snapshot_class:
            raise SnapshotError("Snapshot class is not set.")

        snapshot = self.snapshot_class.create(self)
        return snapshot

    def revert_to_snapshot(self, snapshot):
        """
        Reverts state to the given partial state.

        Parameters
        -----------
        snapshot: Snapshot
            The saved data from a prior pool state.
        """
        snapshot.restore(self)

    @contextmanager
    def use_snapshot_context(self):
        """
        This context manager allows creating and reverting
        to snapshots easily with the syntax:

        with pool.use_snapshot_context() as snapshot:
            pool.trade(i, j, dx)
            ...
            etc.

        The pool state will be reverted after the `with` block
        to the state prior to the block.

        `as snapshot` can be omitted but is handy if you need to
        log or introspect on the state.
        """
        snapshot = self.get_snapshot()
        try:
            yield snapshot
        finally:
            self.revert_to_snapshot(snapshot)


class Snapshot(ABC):
    """
    This class allows customization of snapshot logic, i.e.
    controls how partial states are produced and restored.
    """

    @classmethod
    @abstractmethod
    def create(cls, pool):
        """
        Create a snapshot of the pool's state.

        Parameters
        -----------
        pool: object
            The object whose state we are saving as a snapshot.

        Returns
        -------
        Snapshot
            The saved data from the pool state.
        """
        raise NotImplementedError

    @abstractmethod
    def restore(self, pool):
        """
        Update the pool's state using the snapshot data.

        Parameters
        -----------
        pool: object
            The object whose state we have saved as a snapshot.
        """
        raise NotImplementedError


class CurvePoolBalanceSnapshot(Snapshot):
    """Snapshot that saves pool balances and admin balances."""

    def __init__(self, balances, admin_balances):
        self.balances = balances
        self.admin_balances = admin_balances

    @classmethod
    def create(cls, pool):
        balances = pool.balances.copy()
        admin_balances = pool.admin_balances.copy()
        return cls(balances, admin_balances)

    def restore(self, pool):
        pool.balances = self.balances.copy()
        pool.admin_balances = self.admin_balances.copy()


class CurveMetaPoolBalanceSnapshot(Snapshot):
    """
    Snapshot that saves pool balances and admin balances
    and also the basepool balances and LP total supply.
    """

    def __init__(
        self,
        *,
        balances,
        admin_balances,
        bp_balances,
        bp_admin_balances,
        bp_tokens,
    ):
        self.balances = balances
        self.admin_balances = admin_balances
        self.bp_balances = bp_balances
        self.bp_admin_balances = bp_admin_balances
        self.bp_tokens = bp_tokens

    @classmethod
    def create(cls, pool):
        balances = pool.balances.copy()
        admin_balances = pool.admin_balances.copy()
        basepool = pool.basepool
        bp_balances = basepool.balances.copy()
        bp_admin_balances = basepool.admin_balances.copy()
        bp_tokens = basepool.tokens
        return cls(
            balances=balances,
            admin_balances=admin_balances,
            bp_balances=bp_balances,
            bp_admin_balances=bp_admin_balances,
            bp_tokens=bp_tokens,
        )

    def restore(self, pool):
        pool.balances = self.balances.copy()
        pool.admin_balances = self.admin_balances.copy()
        basepool = pool.basepool
        basepool.balances = self.bp_balances.copy()
        basepool.admin_balances = self.bp_admin_balances.copy()
        basepool.tokens = self.bp_tokens


class CurveCryptoPoolBalanceSnapshot(Snapshot):
    """Snapshot that saves pool balances and admin balances."""

    def __init__(
        self,
        balances,
        D,
        price_scale,
        virtual_price,
        xcp_profit,
        last_prices,
        last_prices_timestamp,
    ):
        self.balances = balances
        self.D = D
        self.price_scale = price_scale
        self.virtual_price = virtual_price
        self.xcp_profit = xcp_profit
        self.last_prices = last_prices
        self.last_prices_timestamp = last_prices_timestamp

    @classmethod
    def create(cls, pool):
        balances = pool.balances.copy()
        D = pool.D
        price_scale = pool.price_scale.copy()
        virtual_price = pool.virtual_price
        xcp_profit = pool.xcp_profit
        last_prices = pool.last_prices.copy()
        last_prices_timestamp = pool.last_prices_timestamp
        return cls(
            balances,
            D,
            price_scale,
            virtual_price,
            xcp_profit,
            last_prices,
            last_prices_timestamp,
        )

    def restore(self, pool):
        pool.balances = self.balances.copy()
        pool.D = self.D
        pool.price_scale = self.price_scale.copy()
        pool.virtual_price = self.virtual_price
        pool.xcp_profit = self.xcp_profit
        pool.last_prices = self.last_prices.copy()
        pool.last_prices_timestamp = self.last_prices_timestamp



"""
Mainly a module to house the `CurveCryptoPool`, a cryptoswap implementation in Python.
"""
import time
from math import isqrt, prod
from typing import List

from curvesim.exceptions import CalculationError, CryptoPoolError
from curvesim.logging import get_logger
from curvesim.pool.base import Pool
from curvesim.pool.snapshot import CurveCryptoPoolBalanceSnapshot

from .calcs import (
    factory_2_coin,
    geometric_mean,
    get_alpha,
    get_p,
    get_y,
    halfpow,
    newton_D,
)

logger = get_logger(__name__)

NOISE_FEE = 10**5  # 0.1 bps
EXP_PRECISION = 10**10
PRECISION = 10**18


class CurveCryptoPool(Pool):  # pylint: disable=too-many-instance-attributes
    """Cryptoswap implementation in Python."""

    snapshot_class = CurveCryptoPoolBalanceSnapshot

    __slots__ = (
        "A",
        "gamma",
        "n",
        "precisions",
        "mid_fee",
        "out_fee",
        "allowed_extra_profit",
        "fee_gamma",
        "adjustment_step",
        "admin_fee",
        "ma_half_time",
        "price_scale",
        "_price_oracle",
        "last_prices",
        "last_prices_timestamp",
        "_block_timestamp",
        "balances",
        "D",
        "virtual_price",
        "tokens",
        "xcp_profit",
        "xcp_profit_a",
        "not_adjusted",
    )

    def __init__(  # pylint: disable=too-many-locals,too-many-arguments
        self,
        A: int,
        gamma: int,
        n: int,
        precisions: List[int],
        mid_fee: int,
        out_fee: int,
        allowed_extra_profit: int,
        fee_gamma: int,
        adjustment_step: int,
        ma_half_time: int,
        price_scale: List[int],
        price_oracle=None,
        last_prices=None,
        last_prices_timestamp=None,
        balances=None,
        D=None,
        tokens=None,
        admin_fee: int = 5 * 10**9,
        xcp_profit=10**18,
        xcp_profit_a=10**18,
    ):
        """
        Parameters
        ----------
        A : int
            Amplification coefficient; this is :math:`A n^n` in the whitepaper
            multiplied by 10**4 for greater precision.
        gamma: int
            Decay factor for A.
        n: int
            Number of coins; currently only n = 2 is supported.
        precisions: list of int
            Precision adjustments to convert native token units to 18 decimals;
            this assumes tokens have at most 18 decimals
            i.e. balance in native units * precision = balance in D units
        mid_fee: int
            Fee with 10**10 precision, for trades near price scale
        out_fee: int
            Fee with 10**10 precision, used to adjust `mid_fee` for trades
            further away from price_scale
        allowed_extra_profit: int
            "Buffer" used to determine if the price adjustment algorithm
            should run.
        fee_gamma: int
            Factor used to control the transition from `mid_fee` to `out_fee`.
        adjustment_step:
            Minimum step size to adjust the price scale.
        ma_half_time: int
            "Half-life" for exponential moving average of trade prices.
        price_scale: List[Int]
            Price scale value for the pool.  This is where liquidity is concentrated.
        price_oracle: List[Int], optional
            Price oracle value for the pool.  This is the EMA price used to
            adjust the price scale toward.
            Defaults to `price_scale`.
        last_prices: List[Int], optional
            Last trade price for the pool.
            Defaults to `price_scale`.
        last_prices_timestamp: int, optional
            Timestamp for last operation altering pool price.
            Defaults to unix timestamp.
        balances: list of int, optional
            Coin balances in native token units;
            either `balances` or `D` is required
        D : int, optional
            Stableswap invariant for given balances, precisions, price_scale,
            A, and gamma; either `balances` or `D` is required
        tokens: int, optional
            LP token supply (default is calculated from `D`, which is also
            calculated if needed)
        admin_fee: int, optional
            Percentage of `fee` with 10**10 precision.  Fee paid to the DAO
            (default = 5*10**9)
        xcp_profit: int, optional
            Counter for accumulated profits, no losses (default = 10**18)
        xcp_profit_a: int, optional
            Value of `xcp_profit` when admin fees last claimed (default = 10**18)
        """
        self.A = A
        self.gamma = gamma

        self.mid_fee = mid_fee
        self.out_fee = out_fee
        self.allowed_extra_profit = allowed_extra_profit
        self.fee_gamma = fee_gamma
        self.adjustment_step = adjustment_step
        self.admin_fee = admin_fee

        self.price_scale = price_scale
        self._price_oracle = price_oracle or price_scale
        self.last_prices = last_prices or price_scale
        self.ma_half_time = ma_half_time

        self._block_timestamp = _get_unix_timestamp()
        self.last_prices_timestamp = last_prices_timestamp or 0

        self.xcp_profit = xcp_profit
        self.xcp_profit_a = xcp_profit_a  # Full profit at last claim of admin fees
        self.not_adjusted = False

        if n not in [2, 3]:
            raise CryptoPoolError(
                "Only 2 or 3-coin crypto pools are currently supported."
            )
        self.n = n
        self.precisions = precisions

        if len(precisions) != n:
            raise ValueError("`len(precisions)` must equal `n`")

        if balances is None and D is None:
            raise ValueError("Must provide at least one of `balances` or `D`.")

        # All state variables needed for balance conversions or "newton"
        # calculations should have been set by this point.

        if balances:
            self.balances = balances

        if D is not None:
            self.D = D
            if not balances:
                self.balances = self._convert_D_to_balances(D)
            else:
                # If user passes both `D` and `balances`, it's possible they may
                # be inconsistent; however we allow this for unanticipated use-cases.
                logger.debug(
                    "Both `D` and `balances` were passed into `__init__`. "
                    "Inconsistent values may create issues."
                )
        else:
            xp = self._xp()
            D = factory_2_coin.newton_D(A, gamma, xp)
            self.D = D

        xcp = self._get_xcp(D)
        if tokens is not None:
            self.tokens = tokens
        else:
            self.tokens = xcp

        self.virtual_price = 10**18 * xcp // tokens

    def _convert_D_to_balances(self, D):
        price_scale = self.price_scale
        precisions = self.precisions
        n = self.n

        return [D // n // precisions[0]] + [
            D * PRECISION // (p * n) // prec
            for p, prec in zip(price_scale, precisions[1:])
        ]

    def _xp(self) -> List[int]:
        """
        Calculate the balances in units of `D`, converting using `price_scale`
        so a unit of each token has equal value.

        Returns
        --------
        List[int]
            The pool balances in units of `D`.

        Note
        -----
        This intentionally always return a new copy of the balances.
        """
        balances = self.balances
        return self._xp_mem(balances)

    def _xp_mem(self, balances) -> List[int]:
        """
        Parameters
        ----------
        balances: List[int]
            The pool balances in native token units.

        Returns
        --------
        List[int]
            The pool balances in units of `D`.

        Note
        -----
        This intentionally always return a new copy of the balances.
        """
        precisions = self.precisions
        price_scale = self.price_scale
        return [balances[0] * precisions[0]] + [
            balance * precision * price // PRECISION
            for balance, precision, price in zip(
                balances[1:], precisions[1:], price_scale
            )
        ]

    def _get_xcp(self, D: int) -> int:
        """
        Calculate the constant-product profit, using the balances at
        equilibrium point.
        """
        n_coins: int = self.n
        price_scale: List[int] = self.price_scale
        x = [D // n_coins] + [
            D * PRECISION // (price * n_coins) for price in price_scale
        ]
        return geometric_mean(x)

    def _increment_timestamp(self, blocks=1, timestamp=None):
        """Update the internal clock used to mimic the block timestamp."""
        if timestamp:
            self._block_timestamp = timestamp
            return

        self._block_timestamp += 12 * blocks

    # pylint: disable-next=R0912,R0913,R0914,R0915
    def _tweak_price(  # noqa: complexity: 12
        self,
        A: int,
        gamma: int,
        _xp: List[int],
        i: int,
        p_i: int,
        new_D: int,
        K0_prev: int = 0,
    ):
        """
        Applies several kinds of updates:
            - EMA price update: price_oracle
            - Profit counters: D, virtual_price, xcp_profit
            - price adjustment: price_scale

        Also claims admin fees if appropriate (enough profit and price scale
        and oracle is close enough).
        """

        price_oracle: List[int] = self._price_oracle
        last_prices: List[int] = self.last_prices
        price_scale: List[int] = self.price_scale
        last_prices_timestamp: int = self.last_prices_timestamp
        block_timestamp: int = self._block_timestamp
        n_coins: int = self.n

        # Update EMA price oracle for a new block.  Happens once per block.
        # EMA uses price of the last trade and oracle price in previous block.
        if last_prices_timestamp < block_timestamp:
            ma_half_time: int = self.ma_half_time
            alpha: int = get_alpha(
                ma_half_time, block_timestamp, last_prices_timestamp, n_coins
            )
            if n_coins == 2:
                price_oracle = [
                    (last_p * (10**18 - alpha) + oracle_p * alpha) // 10**18
                    for last_p, oracle_p in zip(last_prices, price_oracle)
                ]
            elif n_coins == 3:
                # Cap state price that goes into the EMA with 2*price_scale.
                price_oracle = [
                    (min(last_p, 2 * p) * (10**18 - alpha) + oracle_p * alpha)
                    // 10**18
                    for last_p, oracle_p, p in zip(
                        last_prices, price_oracle, price_scale
                    )
                ]
            else:
                raise CalculationError("More than 3 coins is not supported.")
            self._price_oracle = price_oracle
            self.last_prices_timestamp = block_timestamp

        D_unadjusted: int = new_D  # Withdrawal methods know new D already
        if new_D == 0:
            D_unadjusted = newton_D(A, gamma, _xp, K0_prev)

        if p_i > 0:
            # Save the last price
            if i > 0:
                last_prices[i - 1] = p_i
            else:
                # If 0th price changed - change all prices instead
                for k in range(n_coins - 1):
                    last_prices[k] = last_prices[k] * 10**18 // p_i
        else:
            if n_coins == 2:
                # calculate real prices
                __xp: List[int] = _xp.copy()
                dx_price: int = __xp[0] // 10**6
                __xp[0] += dx_price
                last_prices = [
                    price_scale[k - 1]
                    * dx_price
                    // (
                        __xp[k]
                        - factory_2_coin.newton_y(A, gamma, __xp, D_unadjusted, k)
                    )
                    for k in range(1, n_coins)
                ]
            else:
                last_prices = get_p(_xp, D_unadjusted, A, gamma)
                last_prices = [
                    last_p * p // 10**18
                    for last_p, p in zip(last_prices, price_scale)
                ]

        self.last_prices = last_prices

        total_supply: int = self.tokens
        old_xcp_profit: int = self.xcp_profit
        old_virtual_price: int = self.virtual_price

        # Update profit numbers without price adjustment first
        xp: List[int] = [D_unadjusted // n_coins] + [
            D_unadjusted * PRECISION // (n_coins * price) for price in price_scale
        ]
        xcp_profit: int = 10**18
        virtual_price: int = 10**18

        if old_virtual_price > 0:
            xcp: int = geometric_mean(xp)
            virtual_price = 10**18 * xcp // total_supply

            if virtual_price < old_virtual_price:
                raise CryptoPoolError("Loss")

            xcp_profit = old_xcp_profit * virtual_price // old_virtual_price

        self.xcp_profit = xcp_profit

        if virtual_price * 2 - 10**18 > xcp_profit + 2 * self.allowed_extra_profit:
            norm: int = 0
            ratio: int = 0
            for k in range(n_coins - 1):
                ratio = price_oracle[k] * 10**18 // price_scale[k]
                ratio = abs(ratio - 10**18)
                norm += ratio**2
            norm = isqrt(norm)
            adjustment_step: int = max(self.adjustment_step, norm // 5)

            if norm > adjustment_step:
                new_prices = [
                    (p * (norm - adjustment_step) + adjustment_step * p_oracle) // norm
                    for p, p_oracle in zip(price_scale, price_oracle)
                ]

                # Calculate balances * prices
                xp = [_xp[0]] + [
                    balance * p_new // p
                    for balance, p, p_new in zip(_xp[1:], price_scale, new_prices)
                ]

                # Calculate "extended constant product" invariant xCP and virtual price
                D: int = newton_D(A, gamma, xp)
                xp = [D // n_coins] + [
                    D * PRECISION // (n_coins * p_new) for p_new in new_prices
                ]
                new_virtual_price = 10**18 * geometric_mean(xp) // total_supply

                # Proceed if we've got enough profit:
                #   new_virtual_price > 10**18
                #   new_virtual_price - 10**18 > (xcp_profit - 10**18) / 2
                if (new_virtual_price > 10**18) and (
                    2 * new_virtual_price - 10**18 > xcp_profit
                ):
                    self.price_scale = new_prices
                    self.D = D
                    self.virtual_price = new_virtual_price
                    return

        # If we are here, the price_scale adjustment did not happen
        # Still need to update the profit counter and D
        self.D = D_unadjusted
        self.virtual_price = virtual_price

    def _claim_admin_fees(self):
        # no gulping logic needed for the python code
        xcp_profit: int = self.xcp_profit
        xcp_profit_a: int = self.xcp_profit_a

        vprice: int = self.virtual_price

        if xcp_profit > xcp_profit_a:
            fees: int = (xcp_profit - xcp_profit_a) * self.admin_fee // (2 * 10**10)
            if fees > 0:
                frac: int = vprice * 10**18 // (vprice - fees) - 10**18
                d_supply = self.tokens * frac // 10**18
                self.tokens += d_supply
                xcp_profit -= fees * 2
                self.xcp_profit = xcp_profit

        A = self.A
        gamma = self.gamma
        totalSupply = self.tokens

        D: int = factory_2_coin.newton_D(A, gamma, self._xp())
        self.D = D
        self.virtual_price = 10**18 * self._get_xcp(D) // totalSupply

        if xcp_profit > xcp_profit_a:
            self.xcp_profit_a = xcp_profit

    def get_dy(self, i: int, j: int, dx: int) -> int:
        """
        Calculate the amount received from swapping `dx`
        amount of the `i`-th coin for the `j`-th coin.

        Parameters
        ----------
        i: int
            Index of 'in' coin
        j: int
            Index of 'out' coin
        dx: int
            Amount of 'in' coin
        Returns
        -------
        int
            The 'out' coin amount

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        assert i != j  # dev: same input and output coin
        assert i < self.n  # dev: coin index out of range
        assert j < self.n  # dev: coin index out of range

        xp: List[int] = self.balances.copy()
        xp[i] += dx
        xp = self._xp_mem(xp)

        A = self.A
        gamma = self.gamma
        D: int = self.D

        y: int = factory_2_coin.newton_y(A, gamma, xp, D, j)
        dy: int = xp[j] - y - 1
        xp[j] = y
        precisions: List[int] = self.precisions
        price_scale: List[int] = self.price_scale
        if j > 0:
            dy = dy * PRECISION // (price_scale[j - 1] * precisions[j])
        else:
            dy = dy // precisions[0]
        dy -= self._fee(xp) * dy // 10**10

        return dy

    def get_y(self, i, j, x, xp):
        r"""
        Calculate x[j] if one makes x[i] = x.

        Parameters
        ----------
        i: int
            index of coin; usually the "in"-token
        j: int
            index of coin; usually the "out"-token
        x: int
            balance of i-th coin in units of D
        xp: list of int
            coin balances in units of D

        Returns
        -------
        int
            The balance of the j-th coin, in units of D, for the other
            coin balances given.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        A: int = self.A
        gamma: int = self.gamma
        D: int = newton_D(A, gamma, xp)

        xp = xp.copy()
        xp[i] = x

        y, _ = get_y(A, gamma, xp, D, j)
        return y

    def _fee(self, xp: List[int]) -> int:
        """
        f = fee_gamma / (fee_gamma + (1 - K))
        where
        K = prod(x) / (sum(x) / N)**N
        (all normalized to 1e18)
        """
        n_coins: int = self.n
        fee_gamma: int = self.fee_gamma
        if n_coins == 2:
            f: int = xp[0] + xp[1]
            f = (
                fee_gamma
                * 10**18
                // (
                    fee_gamma
                    + 10**18
                    - (10**18 * n_coins**n_coins) * xp[0] // f * xp[1] // f
                )
            )
        else:
            _sum_xp: int = sum(xp)
            K = 10**18
            for _x in xp:
                K = K * n_coins * _x // _sum_xp
            f = fee_gamma * 10**18 // (fee_gamma + 10**18 - K)
        return (self.mid_fee * f + self.out_fee * (10**18 - f)) // 10**18

    # pylint: disable-next=too-many-locals
    def _exchange(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int,
    ) -> int:
        assert i != j, "Indices must be different"
        assert i < self.n, "Index out of bounds"
        assert j < self.n, "Index out of bounds"
        assert dx > 0, "Can't swap zero amount"

        A = self.A
        gamma = self.gamma
        xp: List[int] = self.balances.copy()
        ix: int = j

        y: int = xp[j]
        xp[i] += dx
        self.balances[i] = xp[i]

        xp = self._xp_mem(xp)

        y_out = get_y(A, gamma, xp, self.D, j)
        dy: int = xp[j] - y_out[0]
        assert dy >= 0, f"Invalid dy: dx: {dx}, dy: {dy}, i: {i}, j: {j} "
        xp[j] -= dy
        dy -= 1

        price_scale: int = self.price_scale[j - 1]
        prec_i: int = self.precisions[i]
        prec_j: int = self.precisions[j]

        if j > 0:
            dy = dy * PRECISION // (price_scale)
        dy = dy // prec_j

        fee = self._fee(xp) * dy // 10**10
        dy -= fee
        assert dy >= min_dy, f"Slippage: dy: {dy}"
        y -= dy

        self.balances[j] = y

        y *= prec_j
        if j > 0:
            y = y * price_scale // PRECISION
        xp[j] = y

        p: int = 0
        K0_prev: int = 0
        if self.n == 2:
            if dx > 10**5 and dy > 10**5:
                _dx: int = dx * prec_i
                _dy: int = dy * prec_j
                if i != 0 and j != 0:
                    p = self.last_prices[i - 1] * _dx // _dy
                elif i == 0:
                    p = _dx * 10**18 // _dy
                else:  # j == 0
                    p = _dy * 10**18 // _dx
                    ix = i
        else:
            K0_prev = y_out[1]

        self._tweak_price(A, gamma, xp, ix, p, 0, K0_prev)

        return dy, fee

    def exchange(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int = 0,
    ) -> int:
        """
        Swap `dx` amount of the `i`-th coin for the `j`-th coin.

        Parameters
        ----------
        i: int
            'In' coin index
        j: int
            'Out' coin index
        dx: int
            'In' coin amount
        min_dy: int, optional
            Minimum 'out' coin amount required (default = 0)

        Returns
        -------
        (int, int)
            (amount of coin `j` received, trading fee)

        Note
        -----
        In the vyper contract, there is an option to exchange using WETH or ETH.
        """
        return self._exchange(i, j, dx, min_dy)

    def exchange_underlying(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int = 0,
    ) -> int:
        """
        In the vyper contract, this exchanges using ETH instead of WETH.
        In Curvesim, this is the same as `exchange`.
        """
        return self.exchange(i, j, dx, min_dy)

    # pylint: disable-next=too-many-locals
    def add_liquidity(
        self,
        amounts: List[int],
        min_mint_amount: int = 0,
    ) -> int:
        """
        Add liquidity into the pool by depositing coins for LP tokens.

        Parameters
        ----------
        amounts: List[int]
            Deposit amounts.  At least one coin amount must be nonzero.
        min_mint_amount: int
            Minimum amount of LP tokens required (default = 0)

        Returns
        -------
        int
            Amount of LP tokens minted.
        """
        assert amounts[0] > 0 or amounts[1] > 0  # dev: no coins to add

        A = self.A
        gamma = self.gamma
        n_coins: int = self.n

        xp_old: List[int] = self._xp_mem(self.balances)

        for i in range(n_coins):
            self.balances[i] += amounts[i]

        xp: List[int] = self._xp_mem(self.balances)
        amountsp: List[int] = [xp[i] - xp_old[i] for i in range(n_coins)]

        old_D: int = self.D
        D: int = factory_2_coin.newton_D(A, gamma, xp)

        d_token: int = 0
        token_supply: int = self.tokens
        if old_D > 0:
            d_token = token_supply * D // old_D - token_supply
        else:
            # sets initial virtual price to 1
            d_token = self._get_xcp(D)
        assert d_token > 0  # dev: nothing minted

        d_token_fee: int = 0
        if old_D > 0:
            d_token_fee = self._calc_token_fee(amountsp, xp) * d_token // 10**10 + 1
            d_token -= d_token_fee
            token_supply += d_token
            self.tokens += d_token

            # Calculate price:
            # p_i * (dx_i - dtoken / token_supply * xx_i)
            # = sum{k!=i}(p_k * (dtoken / token_supply * xx_k - dx_k))
            # only ix is nonzero
            p: int = 0
            ix: int = -1
            if d_token > 10**5:
                nonzero_indices = [i for i, a in enumerate(amounts) if a != 0]
                if len(nonzero_indices) == 1:
                    # not reached in current tests, which never have nonzero amounts
                    prec: List[int] = self.precisions
                    last_prices: List[int] = self.last_prices
                    balances: List[int] = self.balances

                    ix = amounts.index(0)
                    S: int = 0
                    for i in range(n_coins):
                        if i == ix:
                            continue

                        if i == 0:
                            S += balances[i] * prec[i]
                        else:
                            S += balances[i] * prec[i] * last_prices[i - 1] // PRECISION
                    S = S * d_token // token_supply
                    p = (
                        S
                        * PRECISION
                        // (
                            amounts[ix] * prec[ix]
                            - d_token * balances[ix] * prec[ix] // token_supply
                        )
                    )

            self._tweak_price(A, gamma, xp, ix, p, D)

        else:
            self.D = D
            self.virtual_price = 10**18
            self.xcp_profit = 10**18
            self.tokens += d_token

        assert d_token >= min_mint_amount, "Slippage"

        return d_token

    def _calc_token_fee(self, amounts: List[int], xp: List[int]) -> int:
        n_coins: int = self.n
        # fee = sum(amounts_i - avg(amounts)) * fee' / sum(amounts)
        fee: int = self._fee(xp) * n_coins // (4 * (n_coins - 1))
        S: int = sum(amounts)
        avg: int = S // n_coins
        Sdiff: int = sum(abs(_x - avg) for _x in amounts)
        return fee * Sdiff // S + NOISE_FEE

    def remove_liquidity(
        self,
        _amount: int,
        min_amounts=None,
    ):
        """
        Remove liquidity (burn LP tokens) to receive back part (or all) of
        the deposited funds.

        Parameters
        ----------
        _amount: int
            Amount LP tokens to burn.
        min_amounts: List[int], optional
            Minimum required amounts for each coin.  Default is 0 each.

        Note
        ----
        "This withdrawal method is very safe, does no complex math"
        """
        min_amounts = min_amounts or [0, 0]

        total_supply: int = self.tokens
        self.tokens -= _amount
        balances: List[int] = self.balances
        amount: int = _amount - 1  # Make rounding errors favoring other LPs a tiny bit

        for i in range(self.n):
            d_balance: int = balances[i] * amount // total_supply
            assert d_balance >= min_amounts[i]
            self.balances[i] = balances[i] - d_balance

        D: int = self.D
        self.D = D - D * amount // total_supply

    def remove_liquidity_one_coin(
        self, token_amount: int, i: int, min_amount: int
    ) -> int:
        """
        Remove liquidity entirely in one type of coin.
        Fees will be extracted and there may be significant price impact incurred.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to burn.
        i: int
            Index of the `out` coin.
        min_amount: int
            Minimum amount of the 'out' coin required (default = 0)

        Returns
        -------
        int
            Amount of the `i`-th coin received.
        """
        A = self.A
        gamma = self.gamma

        dy: int = 0
        D: int = 0
        p: int = 0
        xp = [0] * self.n
        dy, p, D, xp = self._calc_withdraw_one_coin(
            A, gamma, token_amount, i, False, True
        )
        assert dy >= min_amount, "Slippage"

        self.balances[i] -= dy
        self.tokens -= token_amount

        self._tweak_price(A, gamma, xp, i, p, D)

        return dy

    # pylint: disable-next=too-many-locals,too-many-arguments
    def _calc_withdraw_one_coin(
        self,
        A: int,
        gamma: int,
        token_amount: int,
        i: int,
        update_D: bool,
        calc_price: bool,
    ) -> (int, int, int, List[int]):
        token_supply: int = self.tokens
        assert token_amount <= token_supply  # dev: token amount more than supply
        assert i < self.n  # dev: coin out of range

        xx: List[int] = self.balances.copy()
        D0: int = 0
        precisions: List[int] = self.precisions

        xp: List[int] = self._xp_mem(xx)

        if update_D:
            D0 = factory_2_coin.newton_D(A, gamma, xp)
        else:
            D0 = self.D

        D: int = D0

        # Charge fee on D, not on y, e.g. reducing invariant LESS than charging user
        fee: int = self._fee(xp)
        dD: int = token_amount * D // token_supply
        D -= dD - (fee * dD // (2 * 10**10) + 1)
        y: int = factory_2_coin.newton_y(A, gamma, xp, D, i)
        if i == 0:
            dy: int = (xp[i] - y) // precisions[i]
        else:
            dy: int = (
                (xp[i] - y) * PRECISION // (precisions[i] * self.price_scale[i - 1])
            )
        xp[i] = y

        # FIXME: update for n coins
        # Price calc
        p: int = 0
        if calc_price and dy > 10**5 and token_amount > 10**5:
            # p_i = dD / D0 * sum'(p_k * x_k) / (dy - dD / D0 * y0)
            S: int = 0
            precision: int = precisions[0]
            if i == 1:
                S = xx[0] * precisions[0]
                precision = precisions[1]
            else:
                S = xx[1] * precisions[1]
            S = S * dD // D0
            p = S * PRECISION // (dy * precision - dD * xx[i] * precision // D0)
            if i == 0:
                p = (10**18) ** 2 // p

        return dy, p, D, xp

    def calc_withdraw_one_coin(self, token_amount: int, i: int) -> int:
        """
        Calculate the output amount from burning `token amount` of LP tokens
        and receiving entirely in the `i`-th coin.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to burn.
        i: int
            Index of the `out` coin.

        Returns
        -------
        int
            Output amount of the `i`-th coin.
        """
        return self._calc_withdraw_one_coin(
            self.A, self.gamma, token_amount, i, True, False
        )[0]

    def lp_price(self) -> int:
        """
        Returns an LP token price approximating behavior as a constant-product AMM.
        """
        if self.n == 2:
            virtual_price = self.virtual_price
            price_oracle = self.internal_price_oracle()
            price = factory_2_coin.lp_price(virtual_price, price_oracle)
        # TODO: find/implement integer cube root function
        # elif self.n == 3:
        #     price_oracle = self.internal_price_oracle()
        #     price =  (
        #         3 * self.virtual_price * icbrt(price_oracle[0] * price_oracle[1])
        #     ) // 10**24
        else:
            raise CalculationError("LP price calc doesn't support more than 3 coins")

        return price

    def internal_price_oracle(self) -> List[int]:
        """
        Return the value of the EMA price oracle.
        """
        price_oracle: int = self._price_oracle
        last_prices_timestamp: int = self.last_prices_timestamp

        block_timestamp: int = self._block_timestamp
        if last_prices_timestamp < block_timestamp:
            ma_half_time: int = self.ma_half_time
            last_prices: int = self.last_prices
            alpha: int = halfpow(
                (block_timestamp - last_prices_timestamp) * 10**18 // ma_half_time
            )
            return [
                (last_p * (10**18 - alpha) + oracle_p * alpha) // 10**18
                for last_p, oracle_p in zip(last_prices, price_oracle)
            ]

        return price_oracle

    def price_oracle(self) -> List[int]:
        """
        Return the value of the EMA price oracle.

        Same as `internal_price_oracle`.  Kept for compatability with the
        vyper interface.
        """
        return self.internal_price_oracle()

    def get_virtual_price(self) -> int:
        """
        Return the virtual price of an LP token.
        """
        return 10**18 * self._get_xcp(self.D) // self.tokens

    def calc_token_amount(self, amounts: List[int]) -> int:
        """
        Calculate the amount of LP tokens minted by depositing given amounts.

        Parameters
        ----------
        amounts: List[int]
            Deposit amounts.  At least one coin amount must be nonzero.

        Returns
        -------
        int
            Amount of LP tokens minted.
        """
        token_supply: int = self.tokens
        A = self.A
        gamma = self.gamma
        xp: List[int] = self._xp()
        amountsp = self._xp_mem(amounts)
        D0: int = self.D
        for i, a in enumerate(amountsp):
            xp[i] += a
        D: int = factory_2_coin.newton_D(A, gamma, xp)
        d_token: int = token_supply * D // D0 - token_supply
        d_token -= self._calc_token_fee(amountsp, xp) * d_token // 10**10 + 1
        return d_token

    def dydxfee(self, i, j):
        """
        Returns the spot price of i-th coin quoted in terms of j-th coin,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Trading fees are deducted.

        Parameters
        ----------
        i: int
            Index of coin to be priced; in a swapping context, this is
            the "in"-token.
        j: int
            Index of quote currency; in a swapping context, this is the
            "out"-token.

        Returns
        -------
        float
            Price of i-th coin quoted in j-th coin with fees deducted.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        return self.dydx(i, j, use_fee=True)

    def dydx(self, i, j, use_fee=False):  # pylint: disable=too-many-locals
        """
        Returns the spot price of i-th coin quoted in terms of j-th coin,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Defaults to no fees deducted.

        Parameters
        ----------
        i: int
            Index of coin to be priced; in a swapping context, this is
            the "in"-token.
        j: int
            Index of quote currency; in a swapping context, this is the
            "out"-token.

        Returns
        -------
        float
            Price of i-th coin quoted in j-th coin

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        xp = self._xp()
        x_i = xp[i]
        x_j = xp[j]
        n = len(xp)

        D = self.D
        A = self.A
        A_multiplier = 10**4
        gamma = self.gamma

        K0 = 10**18 * n**n * prod(xp) / D**n

        coeff = A * gamma**2 / (10**18 + gamma - K0) ** 2
        frac = (10**18 + gamma + K0) * (sum(xp) - D) / (10**18 + gamma - K0)
        dydx_top = x_j * (A_multiplier * D + coeff * (x_i + frac))
        dydx_bottom = x_i * (A_multiplier * D + coeff * (x_j + frac))
        dydx = dydx_top / dydx_bottom

        if j > 0:
            price_scale = self.price_scale[j - 1]
            dydx = dydx * 10**18 / price_scale
        if i > 0:
            price_scale = self.price_scale[i - 1]
            dydx = dydx * price_scale / 10**18

        if use_fee:
            fee = self._fee(xp)
            dydx = dydx - dydx * fee / 10**10

        return dydx


def _get_unix_timestamp():
    """Get the timestamp in Unix time."""
    return int(time.time())



"""
Iterators for use in simulation runs.
These are input directly into :func:`.pipelines.templates.run_pipeline`,
along with a strategy that is run at each timestemp.

Iterators fall into two general categories:

    1. :mod:`.param_samplers`:
        Generate pools with updated parameters per run.
    2. :mod:`.price_samplers`:
        Generate price, volume, and/or other time-series data per tick.
"""



"""
General utility for http requests.
"""
import aiohttp
from aiohttp import ClientResponseError
from tenacity import retry, stop_after_attempt, wait_exponential

from curvesim.exceptions import HttpClientError

stop_rule = stop_after_attempt(8)
wait_rule = wait_exponential(multiplier=1.5, min=2, max=60)


class HTTP:
    @staticmethod
    @retry(stop=stop_rule, wait=wait_rule)
    async def get(url, params=None):

        kwargs = {"url": url, "headers": {"Accept-Encoding": "gzip"}}

        if params is not None:
            kwargs.update({"params": params})

        try:
            async with aiohttp.request("GET", **kwargs) as resp:
                resp.raise_for_status()
                json_data = await resp.json()
        except ClientResponseError as e:
            message = e.message
            status = e.status
            url = e.request_info.url
            # pylint: disable-next=raise-missing-from
            raise HttpClientError(status, message, url)

        return json_data

    @staticmethod
    @retry(stop=stop_rule, wait=wait_rule)
    async def post(url, json=None):
        kwargs = {"url": url, "headers": {"Accept-Encoding": "gzip"}}

        if json is not None:
            kwargs.update({"json": json})

        try:
            async with aiohttp.request("POST", **kwargs) as resp:
                resp.raise_for_status()
                json_data = await resp.json()
        except ClientResponseError as e:
            message = e.message
            status = e.status
            url = e.request_info.url
            # pylint: disable-next=raise-missing-from
            raise HttpClientError(status, message, url)

        return json_data



"""Base SimPool implementation for Curve stableswap pools, both regular and meta."""
from abc import abstractmethod

from curvesim.exceptions import CurvesimValueError
from curvesim.utils import cache


class AssetIndicesMixin:
    """
    This mixin translates from asset names to Curve pool indices.
    Used in both stableswap and cryptoswap implementations used
    in arbitrage pipelines.
    """

    @property
    @abstractmethod
    def asset_names(self):
        """
        Return list of asset names.

        For metapools, our convention is to place the basepool LP token last.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _asset_balances(self):
        """Return list of asset balances in same order as asset_names."""
        raise NotImplementedError

    @property
    def asset_balances(self):
        """Return dict mapping asset names to coin balances."""
        return dict(zip(self.asset_names, self._asset_balances))

    @property
    @cache
    def asset_indices(self):
        """
        Return dict mapping asset names to pool index.

        For metapools, our convention is to place the basepool LP token last, so that
        underlyer indices are preserved, but the basepool LP token is accessible using
        the index pool.n_total.
        """
        return {name: i for i, name in enumerate(self.asset_names)}

    def get_asset_indices(self, *asset_ids):
        """
        Gets the pool indices for the input asset names.
        Uses the indices set in _asset_indices.
        """
        indices = []
        for ID in asset_ids:
            if isinstance(ID, str):
                ID = self.asset_indices[ID]
            indices.append(ID)

        if len(indices) != len(set(indices)):
            raise CurvesimValueError("Duplicate asset indices.")

        return indices



"""
Implements the volume-limited arbitrage pipeline.
"""

import os

from curvesim.iterators.param_samplers import ParameterizedPoolIterator
from curvesim.iterators.price_samplers import PriceVolume
from curvesim.logging import get_logger
from curvesim.metrics import init_metrics, make_results
from curvesim.pool import get_sim_pool
from curvesim.pool.cryptoswap.pool import CurveCryptoPool
from curvesim.pool_data.cache import PoolDataCache

from .. import run_pipeline
from ..common import DEFAULT_METRICS, DEFAULT_PARAMS, TEST_CRYPTO_PARAMS, TEST_PARAMS
from ..utils import compute_volume_multipliers
from .strategy import VolumeLimitedStrategy

logger = get_logger(__name__)


# pylint: disable-next=too-many-locals
def pipeline(
    pool_metadata,
    pool_data_cache=None,
    *,
    variable_params=None,
    fixed_params=None,
    metrics=None,
    test=False,
    days=60,
    src="coingecko",
    data_dir="data",
    vol_mult=None,
    vol_mode=1,
    ncpu=None,
    end=None,
):
    """
    Implements the volume-limited arbitrage pipeline.

    At each timestep, the pool is arbitraged as close to the prevailing market price
    as possible without surpassing a volume constraint. By default, volume is limited
    to the total market volume at each timestep, scaled by the proportion of
    volume attributable to the pool over the whole simulation period (vol_mult).

    Parameters
    ----------
    pool_metadata : :class:`~curvesim.pool_data.metadata.PoolMetaDataInterface`
        Pool metadata object for the pool of interest.

    variable_params : dict, defaults to broad range of A/fee values
        Pool parameters to vary across simulations.
        keys: pool parameters, values: iterables of ints

        Example
        --------
        >>> variable_params = {"A": [100, 1000], "fee": [10**6, 4*10**6]}

    fixed_params : dict, optional
        Pool parameters set before all simulations.
        keys: pool parameters, values: ints

        Example
        --------
        >>> fixed_params = {"D": 1000000*10**18}

    test : bool, optional
        Overrides variable_params to use four test values:

        .. code-block::

            {"A": [100, 1000], "fee": [3000000, 4000000]}

    days : int, default=60
        Number of days to pull pool and price data for.

    src : str, default="coingecko"
        Source for price/volume data: "coingecko" or "local".

    data_dir : str, default="data"
        relative path to saved price data folder

    vol_mult : dict, default computed from data
        Value(s) multiplied by market volume to specify volume limits
        (overrides vol_mode).

        dict should map from trade-pair tuples to values, e.g.:

        .. code-block::

            {('DAI', 'USDC'): 0.1, ('DAI', 'USDT'): 0.1, ('USDC', 'USDT'): 0.1}


    vol_mode : int, default=1
        Modes for limiting trade volume.

        1: limits trade volumes proportionally to market volume for each pair

        2: limits trade volumes equally across pairs

        3: mode 2 for trades with meta-pool asset, mode 1 for basepool-only trades

    ncpu : int, default=os.cpu_count()
        Number of cores to use.

    Returns
    -------
    SimResults object
    """
    if ncpu is None:
        cpu_count = os.cpu_count()
        ncpu = cpu_count if cpu_count is not None else 1

    fixed_params = fixed_params or {}

    default_params = DEFAULT_PARAMS.copy()
    for key in DEFAULT_PARAMS:
        if key in fixed_params:
            del default_params[key]

    variable_params = variable_params or default_params
    metrics = metrics or DEFAULT_METRICS

    if pool_data_cache is None:
        pool_data_cache = PoolDataCache(pool_metadata, days=days, end=end)

    pool = get_sim_pool(pool_metadata, pool_data_cache=pool_data_cache)

    if test:
        fixed_params = {}
        if isinstance(pool, CurveCryptoPool):
            variable_params = TEST_CRYPTO_PARAMS
        else:
            variable_params = TEST_PARAMS

    # pylint: disable-next=abstract-class-instantiated
    param_sampler = ParameterizedPoolIterator(pool, variable_params, fixed_params)
    price_sampler = PriceVolume(
        pool.assets, days=days, data_dir=data_dir, src=src, end=end
    )

    if vol_mult is None:
        total_pool_volume = pool_data_cache.volume
        total_market_volume = price_sampler.total_volumes()
        vol_mult = compute_volume_multipliers(
            total_pool_volume,
            total_market_volume,
            pool_metadata.n,
            pool_metadata.pool_type,
            mode=vol_mode,
        )

    metrics = init_metrics(metrics, pool=pool)
    strategy = VolumeLimitedStrategy(metrics, vol_mult)

    output = run_pipeline(param_sampler, price_sampler, strategy, ncpu=ncpu)
    results = make_results(*output, metrics)

    return results



from . import pipeline

if __name__ == "__main__":
    pool_address = "0xbebc44782c7db0a1a60cb6fe97d0b483032ff1c7"
    chain = "mainnet"
    results = pipeline(pool_address, chain, test=True, ncpu=1)



"""
This module provides functionality for creating and manipulating chart properties.

It contains utility functions for creating keyword arguments for a chart, updating
and initializing chart properties, and handling property-specific behaviors.
"""
from copy import deepcopy
from functools import partial

from altair import Color, MarkDef, X, Y


def make_chart_kwargs(default, override_dicts):
    """
    Create the keyword arguments for a chart by updating a default dictionary with
    values from a list of override dictionaries.

    Parameters
    ----------
    default : dict
        The default keyword arguments for the chart.
    override_dicts : list
        A list of dictionaries containing addtional keyword arguments and/or overrides
        for the default keyword arguments.

    Returns
    -------
    dict
        A dictionary containing the keyword arguments for the chart.
    """
    chart_kwargs = deepcopy(default)

    for d in override_dicts:
        for key, val in d.items():
            update = UPDATE_RULES.get(key, update_properties)
            update(chart_kwargs, key, val)

    return init_properties(chart_kwargs)


def update_properties(prop_dict, key, val, depth=0, count=0):
    """
    Update properties in a dictionary. If the current depth matches the target depth,
    the value is updated directly. Otherwise, the function is called recursively on
    nested dictionaries.

    Parameters
    ----------
    prop_dict : dict
        The dictionary of properties to update.
    key : str
        The key of the property to update.
    val : any
        The new value for the property.
    depth : int, optional
        The target depth for the update. Defaults to 0.
    count : int, optional
        The current depth of the update. Defaults to 0.
    """

    if depth in [0, count]:
        prop_dict[key] = val
    else:
        prop_dict.setdefault(key, {})
        for k, v in val.items():
            update_properties(prop_dict[key], k, v, depth, count + 1)


def concat_properties(prop_dict, key, val):
    """
    Concatenate values to a property in a dictionary. If the value is a list, extend
    the property with it. Otherwise, append the value to the property.

    Parameters
    ----------
    prop_dict : dict
        The dictionary of properties to update.
    key : str
        The key of the property to update.
    val : any
        The value to concatenate to the property.
    """
    prop_dict.setdefault(key, [])

    if isinstance(val, list):
        prop_dict[key] += val
    else:
        prop_dict[key].append(val)


def ignore_property(prop_dict, key, val):  # pylint: disable=unused-argument
    """
    Ignore a property during an update operation. This function does nothing and is
    intended to be used as a placeholder in UPDATE_RULES where certain properties
    should be ignored.

    Parameters
    ----------
    prop_dict : dict
        The dictionary of properties to update.
    key : str
        The key of the property to ignore.
    val : any
        The value of the property to ignore.
    """


def init_properties(prop_dict, classes=None):
    """
    Initialize the properties of a dictionary using a mapping of classes. If a
    property's key is present in the classes dictionary, its value is replaced
    with an instance of the corresponding class, initialized with the value as
    keyword arguments.

    Parameters
    ----------
    prop_dict : dict
        The dictionary of properties to initialize.
    classes : dict, optional
        A dictionary mapping keys to classes. If a key in prop_dict matches a
        key in classes, the value in prop_dict is replaced with an instance of
        the corresponding class. Defaults to PROPERTY_CLASSES.

    Returns
    -------
    dict
        The dictionary of properties, with values replaced with class instances
        where applicable.
    """
    classes = classes or PROPERTY_CLASSES
    for key, val in prop_dict.items():
        try:
            prop_class = classes[key]
        except KeyError:
            continue

        if isinstance(prop_class, dict):
            prop_dict[key] = init_properties(val, classes=classes[key])
        else:
            prop_dict[key] = prop_class(**val)

    return prop_dict


PROPERTY_CLASSES = {
    "encoding": {
        "x": X,
        "y": Y,
        "color": Color,
    },
    "mark": MarkDef,
}

UPDATE_RULES = {
    "mark": partial(update_properties, depth=1),
    "encoding": partial(update_properties, depth=2),
    "selection": update_properties,
    "transform": concat_properties,
    "resample": ignore_property,
    "style": ignore_property,
}



import pytest

from curvesim.exceptions import UnregisteredPoolError
from curvesim.metrics.state_log.pool_parameters import get_pool_parameters
from curvesim.metrics.state_log.pool_state import get_pool_state

# To Do: Currently only get_pool_parameters/get_pool_state are tested.


def test_get_pool_parameters_curve_pool(sim_curve_pool):
    """Test that get_pool_parameters returns the right data for SimCurvePool."""
    pool = sim_curve_pool

    expected_params = {
        "A": pool.A,
        "D": pool.D() / 10**18,
        "fee": pool.fee / 10**10,
        "admin_fee": pool.admin_fee / 10**10,
    }

    params = get_pool_parameters(pool)
    assert params == expected_params

    # Test with fee_mul
    pool.fee_mul = 2 * 10**10
    expected_params["fee_mul"] = 2
    params = get_pool_parameters(pool)
    assert params == expected_params


def test_get_pool_parameters_curve_crypto_pool(sim_curve_crypto_pool):
    """Test that get_pool_parameters returns the right data for SimCurveCryptoPool."""
    pool = sim_curve_crypto_pool

    expected_params = {
        "A": pool.A,
        "gamma": pool.gamma / 10**18,
        "D": pool.D / 10**18,
        "mid_fee": pool.mid_fee / 10**10,
        "out_fee": pool.out_fee / 10**10,
        "fee_gamma": pool.fee_gamma / 10**18,
        "allowed_extra_profit": pool.allowed_extra_profit / 10**18,
        "adjustment_step": pool.adjustment_step / 10**18,
        "ma_half_time": pool.ma_half_time,
        "admin_fee": pool.admin_fee / 10**10,
    }

    params = get_pool_parameters(pool)
    assert params == expected_params


def test_get_pool_parameters_curve_meta_pool(sim_curve_meta_pool):
    """Test that get_pool_parameters returns the right data for SimCurveMetaPool."""
    _test_get_pool_parameters_curve_meta_pool(sim_curve_meta_pool)


def test_get_pool_parameters_curve_rai_pool(sim_curve_rai_pool):
    """Test that get_pool_parameters returns the right data for SimCurveRaiPool."""
    _test_get_pool_parameters_curve_meta_pool(sim_curve_rai_pool)


def _test_get_pool_parameters_curve_meta_pool(pool):
    expected_params = {
        "A": pool.A,
        "D": pool.D() / 10**18,
        "fee": pool.fee / 10**10,
        "admin_fee": pool.admin_fee / 10**10,
        "A_base": pool.basepool.A,
        "D_base": pool.basepool.D() / 10**18,
        "fee_base": pool.basepool.fee / 10**10,
        "admin_fee_base": pool.basepool.admin_fee / 10**10,
    }

    params = get_pool_parameters(pool)
    assert params == expected_params

    # Test with fee_mul
    pool.fee_mul = 2 * 10**10
    pool.basepool.fee_mul = 3 * 10**10
    expected_params["fee_mul"] = 2
    expected_params["fee_mul_base"] = 3

    params = get_pool_parameters(pool)
    assert params == expected_params


def test_get_pool_parameters_unmapped_exception():
    """Test that unmapped pool throws error."""

    class DummyPool:
        pass

    pool = DummyPool()

    with pytest.raises(UnregisteredPoolError):
        get_pool_parameters(pool)


def test_get_pool_state_curve_pool(sim_curve_pool):
    """Test that get_pool_state returns the right data for SimCurvePool."""
    pool = sim_curve_pool

    expected_state = {
        "balances": pool.balances,
        "tokens": pool.tokens,
        "admin_balances": pool.admin_balances,
    }

    state = get_pool_state(pool)
    assert state == expected_state


def test_get_pool_state_curve_crypto_pool(sim_curve_crypto_pool):
    """Test that get_pool_state returns the right data for SimCurveCryptoPool."""
    pool = sim_curve_crypto_pool

    expected_state = {
        "D": pool.D,
        "balances": pool.balances,
        "tokens": pool.tokens,
        "price_scale": pool.price_scale,
        "_price_oracle": pool._price_oracle,
        "xcp_profit": pool.xcp_profit,
        "xcp_profit_a": pool.xcp_profit_a,
        "last_prices": pool.last_prices,
        "last_prices_timestamp": pool.last_prices_timestamp,
        "not_adjusted": pool.not_adjusted,
    }

    state = get_pool_state(pool)
    assert state == expected_state


def test_get_pool_state_curve_meta_pool(sim_curve_meta_pool):
    """Test that get_pool_state returns the right data for SimCurveMetaPool."""
    _test_get_pool_state_curve_meta_pool(sim_curve_meta_pool)


def test_get_pool_state_curve_rai_pool(sim_curve_rai_pool):
    """Test that get_pool_state returns the right data for SimCurveRaiPool."""
    _test_get_pool_state_curve_meta_pool(sim_curve_rai_pool)


def _test_get_pool_state_curve_meta_pool(pool):
    expected_state = {
        "balances": pool.balances,
        "tokens": pool.tokens,
        "admin_balances": pool.admin_balances,
        "balances_base": pool.basepool.balances,
        "tokens_base": pool.basepool.tokens,
        "admin_balances_base": pool.basepool.admin_balances,
    }

    state = get_pool_state(pool)
    assert state == expected_state


def test_get_pool_state_unmapped_exception():
    """Test that unmapped pool throws StateLogError."""

    class DummyPool:
        pass

    pool = DummyPool()

    with pytest.raises(UnregisteredPoolError):
        get_pool_state(pool)



"""
Functions and interfaces used in the simulation pipeline framework.
"""

__all__ = [
    "Trader",
    "Strategy",
    "SimAssets",
    "SimPool",
    "Trade",
    "TradeResult",
    "ParameterSampler",
    "PriceSample",
    "PriceSampler",
]

from .param_samplers import ParameterSampler
from .price_samplers import PriceSample, PriceSampler
from .sim_assets import SimAssets
from .sim_pool import SimPool
from .strategy import Strategy
from .trader import Trade, Trader, TradeResult



"""
Module housing the required implementation for any pool entity
in the Curvesim Framework.
"""


from curvesim.pool.snapshot import SnapshotMixin


class Pool(SnapshotMixin):
    """
    The `Pool` base class has the explicitly required properties for any
    pool-like object used in Curvesim.

    The `SnapshotMixin` gives the ability to snapshot balances and revert
    balance changes.

    Currently the base attributes are not informative for pools
    constructed manually rather than from chain data.
    """

    # need to configure in derived class otherwise the
    # snapshotting will not work
    snapshot_class = None

    @property
    def name(self):
        """Descriptive name for this pool"""
        if hasattr(self, "metadata"):
            return self.metadata["name"]
        return "Custom pool"

    @property
    def address(self):
        """Address on chain"""
        if hasattr(self, "metadata"):
            return self.metadata["address"]
        return "0x" + "0" * 40

    @property
    def coin_names(self):
        """Symbols for the pool coins."""
        if hasattr(self, "metadata"):
            return self.metadata["coins"]["names"]
        return []

    @property
    def coin_addresses(self):
        """Addresses for the pool coins."""
        if hasattr(self, "metadata"):
            return self.metadata["coins"]["addresses"]
        return []

    @property
    def coin_decimals(self):
        """Addresses for the pool coins."""
        if hasattr(self, "metadata"):
            return self.metadata["coins"]["decimals"]
        return []

    @property
    def chain(self):
        """Chain for this pool"""
        if hasattr(self, "metadata"):
            return self.metadata["chain"]
        return "No chain"

    @property
    def pool_type(self):
        """Type of this pool"""
        if hasattr(self, "metadata"):
            return self.metadata["pool_type"]
        return "Custom"

    @property
    def symbol(self):
        """LP token symbol"""
        if hasattr(self, "metadata"):
            return self.metadata["symbol"]
        return "Custom"

    @property
    def folder_name(self):
        """Name of folder containing saved sim results."""
        if hasattr(self, "metadata"):
            symbol = self.symbol
            address = self.address
            return symbol.lower() + "_" + address[:7].lower()
        return "custom_pool_" + str(hash(self))

    def __repr__(self):
        return f"<{self.__class__.__name__} address={self.address} chain={self.chain}>"



from abc import ABC, abstractmethod

from curvesim.logging import get_logger

logger = get_logger(__name__)


class SimPool(ABC):
    """
    The interface that must be implemented by pools used in simulations.

    See curvesim.pool.sim_interface for implementations.
    """

    def prepare_for_trades(self, timestamp):
        """
        Does any necessary preparation before computing and doing trades.

        The input timestamp can be used to fetch any auxiliary data
        needed to prep the state.

        Base implementation is a no-op.

        Parameters
        ----------
        timestamp : datetime.datetime
            the time to sample from
        """

    def prepare_for_run(self, prices):
        """
        Does any necessary preparation before beginning a simulation run.

        Base implementation is a no-op.

        Parameters
        ----------
        timestamp : pandas.DataFrame
            The price time_series, price_sampler.prices.
        """

    @abstractmethod
    def price(self, coin_in, coin_out, use_fee=True):
        """
        Returns the spot price of `coin_in` quoted in terms of `coin_out`,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Coin IDs should be strings but as a legacy feature integer indices
        corresponding to the pool implementation are allowed (caveat lector).

        The indices are assumed to include base pool underlyer indices.

        Parameters
        ----------
        coin_in : str, int
            ID of coin to be priced; in a swapping context, this is
            the "in"-token.
        coin_out : str, int
            ID of quote currency; in a swapping context, this is the
            "out"-token.
        use_fee: bool, default=True
            Deduct fees.

        Returns
        -------
        float
            Price of `coin_in` quoted in `coin_out`
        """
        raise NotImplementedError

    @abstractmethod
    def trade(self, coin_in, coin_out, size):
        """
        Perform an exchange between two coins.

        Coin IDs should be strings but as a legacy feature integer indices
        corresponding to the pool implementation are allowed (caveat lector).

        Note that all amounts are normalized to be in the same units as
        pool value, e.g. for Curve Stableswap pools, the same units as `D`.
        This simplifies cross-token comparisons and creation of metrics.


        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.
        coin_out : str, int
            ID of "out" coin.
        size : int
            Amount of coin `i` being exchanged.

        Returns
        -------
        (int, int)
            (amount of coin `j` received, trading fee)
        """
        raise NotImplementedError

    @abstractmethod
    def get_max_trade_size(self, coin_in, coin_out, out_balance_perc):
        """
        Calculate the swap amount of the "in" coin needed to leave
        the specified percentage of the "out" coin.

        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.
        coin_out : str, int
            ID of "out" coin.
        out_balance_perc : float
            Percentage of the "out" coin balance that should remain
            after doing the swap.

        Returns
        -------
        int
            The amount of "in" coin needed.
        """
        raise NotImplementedError

    def get_min_trade_size(self, coin_in):
        """
        Return the minimal trade size allowed for the pool.

        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.

        Returns
        -------
        int
            The minimal trade size
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def assets(self):
        """
        Return :class:`.SimAssets` object with the properties of the pool's assets.

        Returns
        -------
        SimAssets
            SimAssets object that stores the properties of the pool's assets.
        """
        raise NotImplementedError



from pandas import concat

from curvesim.plot.altair import result_plotter as altair_plotter


class SimResults:
    """
    Results container with methods to plot or return metrics as DataFrames.
    """

    __slots__ = [
        "data_per_run",
        "data_per_trade",
        "summary_data",
        "factors",
        "plot_config",
        "plotter",
    ]

    def __init__(
        self,
        *,
        data_per_run,
        data_per_trade,
        summary_data,
        factors,
        plot_config,
        plotter=altair_plotter,
    ):
        self.data_per_run = data_per_run
        self.data_per_trade = data_per_trade
        self.summary_data = summary_data

        self.factors = factors
        self.plot_config = plot_config
        self.plotter = plotter

    def summary(self, full=False, columns=None):
        """
        Returns a DataFrame of summary metrics.

        Parameters
        ----------
        full : bool, default=False
            If true, includes per-run data (e.g., pool parameters) in the output.

        columns: list, optional
            The metrics to include in the output DataFrame. Top level metric names
            (e.g., "pool_balance") should be used without specifying the individual
            summary statistics (e.g., not "pool_balance min").

            By default, includes all metrics.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with metrics as columns and each simulation run as rows.

        """
        data = get_columns(self.summary_data, columns)

        if full or columns:
            runs = get_columns(self.data_per_run, columns)
            cols = [" ".join(c) for c in data.columns.values]  # flatten multi-index
            data = data.set_axis(cols, axis=1)
            return concat([runs, data], axis=1)

        return data

    def data(self, full=False, columns=None):
        """
        Returns a DataFrame of metrics for each time-point in the simulation.

        Parameters
        ----------
        full : bool, default=False
            If true, includes per-run data (e.g., pool parameters) in the output.

        columns: list, optional
            The metrics to include in the output DataFrame.
            By default, includes all metrics.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with metrics as columns and each timestamp in each run as rows.

        """
        data = get_columns(self.data_per_trade, columns, defaults=["run", "timestamp"])

        if full or columns:
            runs = get_columns(self.data_per_run, columns)
            runs = runs.loc[data["run"]].reset_index(drop=True)  # include run data
            return concat([runs, data], axis=1)

        return data

    def plot(self, summary=True, data=True, save_as=None):
        """
        Returns and optionally saves a plot of the results data.

        Parameters
        ----------
        summary : bool, default=True
            If true, includes summary data in the plot.

        data : bool, default=True
            If true, includes timeseries data in the plot.

        save_as : str, optional
            Path to save plot output to. Typically an .html file. See
            `Altair docs <https://altair-viz.github.io/user_guide/saving_charts.html>`_
            for additional options.


        Returns
        -------
        altair.VConcatChart

        """
        return self.plotter.plot(self, summary, data, save_as)


def get_columns(data, columns, defaults=None):
    """Returns specified DataFrame columns if present. If none, returns all columns."""

    if not columns:
        return data
    if defaults:
        columns = defaults + columns
    columns = [c for c in columns if c in data.columns]
    return data[columns]



"""Pool fixtures to test against vyper implementation.  Uses titanoboa."""
# pylint: disable=redefined-outer-name
import os

import boa
import pytest

_base_dir = os.path.dirname(__file__)
_curve_dir = os.path.join(_base_dir, "curve")
FAKE_ADDRESS = "0xCAFECAFECAFECAFECAFECAFECAFECAFECAFECAFE"


@pytest.fixture(scope="session")
def mainnet_3pool_state():
    """Snapshot of Mainnet 3Pool values"""
    p = [10**18, 10**30, 10**30]
    balances = [
        295949605740077243186725223,
        284320067518878,
        288200854907854,
    ]
    virtual_balances = [b * p // 10**18 for b, p in zip(balances, p)]

    return {
        "N_COINS": 3,
        "A": 2000,
        "p": p,
        "balances": balances,
        "virtual_balances": virtual_balances,
        "lp_tokens": 849743149250065202008212976,
        "virtual_price": 1022038799187029697,
    }


@pytest.fixture(scope="session")
def _vyper_3pool(mainnet_3pool_state):
    """Initialize vyper fixture using mainnet values."""
    lp_total_supply = mainnet_3pool_state["lp_tokens"]
    mock_filepath = os.path.join(_base_dir, "lp_token_mock.vy")
    lp_token = boa.load(mock_filepath, lp_total_supply)

    pool_filepath = os.path.join(_curve_dir, "basepool.vy")
    owner = FAKE_ADDRESS
    coins = [FAKE_ADDRESS] * 3
    A = mainnet_3pool_state["A"]
    fee = 4 * 10**6
    admin_fee = 5 * 10**9
    pool = boa.load(pool_filepath, owner, coins, lp_token, A, fee, admin_fee)

    balances = mainnet_3pool_state["balances"]
    pool.eval(f"self.balances={balances}")

    return pool


@pytest.fixture(scope="session")
def _vyper_metapool(_vyper_3pool):
    """Initialize vyper fixture using mainnet values."""
    metapool_filepath = os.path.join(_curve_dir, "metapool.vy")
    name = "SIM-3Pool"
    symbol = "SIM3CRV-f"
    coin = FAKE_ADDRESS
    rate_multiplier = 10**34  # 2 decimals
    basepool = _vyper_3pool.address
    basepool_token = _vyper_3pool.token()
    A = 1000
    fee = 4 * 10**6
    # Admin fee is hard-coded as 50% for factory pools
    # admin_fee = 5 * 10**9
    metapool = boa.load(
        metapool_filepath,
        name,
        symbol,
        coin,
        basepool,
        basepool_token,
        rate_multiplier,
        A,
        fee,
    )

    balances = [762951074, 12971664836474542835562756]
    metapool.eval(f"self.balances={balances}")
    total_supply = 20312687702458911532611097
    metapool.eval(f"self.totalSupply={total_supply}")

    return metapool


@pytest.fixture(scope="session")
def _cryptopool_lp_token():
    lp_total_supply = 16060447504694332256465310
    mock_filepath = os.path.join(_base_dir, "lp_token_mock.vy")
    lp_token = boa.load(mock_filepath, lp_total_supply)
    return lp_token


@pytest.fixture(scope="session")
def _vyper_cryptopool(_cryptopool_lp_token):
    """
    Initialize vyper fixture for crypto pool
    using default volatile pair settings
    """
    cryptopool_filepath = os.path.join(_curve_dir, "cryptopool.vy")
    coins = [FAKE_ADDRESS] * 2

    # settings based on STG/USDC pool
    # https://etherscan.io/address/0x3211c6cbef1429da3d0d58494938299c92ad5860

    # lp_total_supply = 16060447504694332256465310
    # mock_filepath = os.path.join(_base_dir, "lp_token_mock.vy")
    # lp_token = boa.load(mock_filepath, lp_total_supply)

    A = 400000
    gamma = 72500000000000
    # unpacked_precisions = [10**0, 10**12]
    precisions = 12 << 8
    precisions = precisions | 0  # for explicitness
    mid_fee = 26000000
    out_fee = 45000000
    allowed_extra_profit = 2000000000000
    fee_gamma = 230000000000000
    adjustment_step = 146000000000000
    admin_fee = 5000000000
    ma_half_time = 600
    initial_price = 1550997347493624157

    cryptopool = boa.load(
        cryptopool_filepath,
        A,
        gamma,
        mid_fee,
        out_fee,
        allowed_extra_profit,
        fee_gamma,
        adjustment_step,
        admin_fee,
        ma_half_time,
        initial_price,
        _cryptopool_lp_token,
        coins,
        precisions,
    )

    balances = [20477317313816545807568241, 13270936465339]
    cryptopool.eval(f"self.balances={balances}")

    D = 41060496962103963853877954
    cryptopool.eval(f"self.D={D}")

    virtual_price = 1026434015737186294
    cryptopool.eval(f"self.virtual_price={virtual_price}")

    xcp_profit = 1052829794354693246
    cryptopool.eval(f"self.xcp_profit={xcp_profit}")

    xcp_profit_a = 1052785575319598710
    cryptopool.eval(f"self.xcp_profit_a={xcp_profit_a}")

    last_prices_timestamp = 1689085619
    cryptopool.eval(f"self.last_prices_timestamp={last_prices_timestamp}")

    return cryptopool


@pytest.fixture(scope="function")
def vyper_3pool(_vyper_3pool):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _vyper_3pool


@pytest.fixture(scope="function")
def vyper_metapool(_vyper_metapool):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _vyper_metapool


@pytest.fixture(scope="function")
def cryptopool_lp_token(_cryptopool_lp_token):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _cryptopool_lp_token


@pytest.fixture(scope="function")
def vyper_cryptopool(_vyper_cryptopool):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _vyper_cryptopool


def unpack_3_uint64s(packed_nums):
    mask = 2**64 - 1
    return [
        (packed_nums >> 128) & mask,
        (packed_nums >> 64) & mask,
        packed_nums & mask,
    ]


def pack_3_uint64s(nums):
    return (nums[0] << 128) | (nums[1] << 64) | nums[0]


def pack_prices(prices):
    return (prices[1] << 128) | prices[0]


@pytest.fixture(scope="session")
def _vyper_tricrypto():
    """
    Initialize vyper fixture for crypto pool
    using default volatile pair settings
    """
    trycrypto_ng_filepath = os.path.join(_curve_dir, "tricrypto_ng.vy")
    tricrypto_math_filepath = os.path.join(_curve_dir, "tricrypto_math.vy")
    coins = [FAKE_ADDRESS] * 3

    # settings based on TricryptoUSDT pool
    # https://etherscan.io/address/0xf5f5b97624542d72a9e06f04804bf81baa15e2b4

    A = 1707629
    gamma = 11809167828997
    packed_A_gamma = (A << 128) | gamma
    assert packed_A_gamma == 581076037942835227425498917514114728328226821

    # unpacked_precisioins = [1000000000000, 10000000000, 1]
    packed_precisions = 1000000000000 << 64
    packed_precisions = (packed_precisions | 10000000000) << 64
    packed_precisions = packed_precisions | 1
    assert packed_precisions == 340282366920938463463559074872505306972160000000001

    mid_fee = 3000000
    out_fee = 30000000
    fee_gamma = 500000000000000
    packed_fee_params = mid_fee << 64
    packed_fee_params = (packed_fee_params | out_fee) << 64
    packed_fee_params = packed_fee_params | fee_gamma

    allowed_extra_profit = 2000000000000
    adjustment_step = 490000000000000
    ma_half_time = 865
    packed_rebalancing_params = allowed_extra_profit << 64
    packed_rebalancing_params = (packed_rebalancing_params | adjustment_step) << 64
    packed_rebalancing_params = packed_rebalancing_params | ma_half_time
    assert (
        packed_rebalancing_params == 680564733841876935965653810981216714752000000000865
    )

    # use current price scale as initial prices to match the balances
    # packing is in reverse order
    # 30468634274925745130207
    # 1877445901676407991006
    packed_prices = 1877445901676407991006 << 128
    packed_prices = packed_prices | 30468634274925745130207

    _name = "TricryptoUSDT"
    _symbol = "crvUSDTWBTCWETH"
    _salt = bytes.fromhex(
        "B90B4B3B1043EAF7E27EF307E5FD67AF029117766661203412EDAB9E18E8F6B3"
    )
    _weth = FAKE_ADDRESS
    _math = boa.load(tricrypto_math_filepath)

    tricrypto = boa.load(
        trycrypto_ng_filepath,
        _name,
        _symbol,
        coins,
        _math.address,
        _weth,
        _salt,
        packed_precisions,
        packed_A_gamma,
        packed_fee_params,
        packed_rebalancing_params,
        packed_prices,
    )
    """
    name :
    TricryptoUSDT
    symbol :
    crvUSDTWBTCWETH
    weth :
    0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2
    coins :
    0xdAC17F958D2ee523a2206206994597C13D831ec7
    0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599
    0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2
    math :
    0xcBFf3004a20dBfE2731543AA38599A526e0fD6eE
    salt :
    B90B4B3B1043EAF7E27EF307E5FD67AF029117766661203412EDAB9E18E8F6B3
    packed_precisions :
    340282366920938463463559074872505306972160000000001
    packed_A_gamma :
    581076037942835227425498917514114728328226821
    packed_fee_params :
    1020847100762815390943526144507091182848000000
    packed_rebalancing_params :
    680564733841876935965653810981216714752000000000865
    packed_prices :
    650960167919755280605435624016972588543318000000000000000000
    """

    balances = [18418434882428, 60547327748, 9914993293693631287774]
    tricrypto.eval(f"self.balances={balances}")

    D = 55481143937271477730517113
    tricrypto.eval(f"self.D={D}")

    virtual_price = 1000225178597346879
    tricrypto.eval(f"self.virtual_price={virtual_price}")

    xcp_profit = 1000448625854298803
    tricrypto.eval(f"self.xcp_profit={xcp_profit}")

    xcp_profit_a = 1000440033249679801
    tricrypto.eval(f"self.xcp_profit_a={xcp_profit_a}")

    lp_total_supply = 47986553926751950746367
    tricrypto.eval(f"self.totalSupply={lp_total_supply}")

    price_oracle = [30435581307494178154980, 1870286625867949317551]
    price_oracle_packed = pack_prices(price_oracle)
    tricrypto.eval(f"self.price_oracle_packed={price_oracle_packed}")

    last_prices = [30453123431671769818574, 1871140849377954208512]
    last_prices_packed = pack_prices(last_prices)
    tricrypto.eval(f"self.last_prices_packed={last_prices_packed}")

    last_prices_timestamp = 1689085619
    tricrypto.eval(f"self.last_prices_timestamp={last_prices_timestamp}")

    return tricrypto


@pytest.fixture(scope="function")
def vyper_tricrypto(_vyper_tricrypto):
    """
    Function-scope fixture using titanoboa's snapshotting
    feature to avoid expensive loading.
    """
    with boa.env.anchor():
        yield _vyper_tricrypto



"""
Network connector for LLama Airforce REST API.
"""
# pylint: disable=redefined-outer-name
from asyncio import gather

from eth_utils import to_checksum_address
from numpy import array

from .http import HTTP
from .utils import sync

URL = "https://api-py.llama.airforce/curve/v1/"


# Query Llama Airforce API
async def llamaAPI(endpoint):
    path = "/"
    path = path.join(endpoint)
    url = URL + path

    r = await HTTP.get(url)

    return r


# Specific Queries
async def pools(chain):
    r = await llamaAPI(["pools", chain, "all"])
    pools = r["pools"]

    return pools


async def pool_metadata(address, chain):
    address = address.lower()
    r = await llamaAPI(["pools", chain, address])
    pool = r["pools"][0]

    return pool


async def symbol_address(symbol, chain):
    pls = await pools(chain)

    for pool in pls:
        if pool["symbol"] == symbol:
            addr = to_checksum_address(pool["address"])
            return addr

    raise ValueError("Address for pool symbol not found")


async def volume(address, chain, bp=None, days=60, inUSD=True):
    address = address.lower()

    key = "volume"
    if inUSD:
        key = key + "USD"

    if bp is None:
        r = await llamaAPI(["pools", chain, "volume", address])

        sum_volume = array(0.0)
        for i in range(days):
            sum_volume += r["volume"][i][key]

    else:
        tasks = [
            llamaAPI(["pools", chain, "volume", address]),
            llamaAPI(["pools", chain, "volume", bp]),
        ]

        r_meta, r_base = await gather(*tasks)

        sum_volume = array([0.0, 0.0])
        for i in range(days):
            sum_volume += [r_meta["volume"][i][key], r_base["volume"][i][key]]

    return sum_volume


async def reserves(address, chain):
    address = address.lower()
    r = await llamaAPI(["pools", chain, "reserves", address])
    reserves = r["reserves"]

    return reserves


async def latest_reserves(address, chain, p=10**18, bp=None, inUSD=True):

    key = "reserves"
    if inUSD:
        key = key + "USD"

    if bp is None:
        res = await reserves(address, chain)
        curr_reserves = res[0][key]
        curr_reserves = [int(r * p) for r in curr_reserves]

    else:
        curr_reserves = await gather(
            latest_reserves(address, chain, p=p, inUSD=inUSD),
            latest_reserves(bp, chain, p=p, inUSD=inUSD),
        )

    return curr_reserves


# Sync
llamaAPI_sync = sync(llamaAPI)
pools_sync = sync(pools)
pool_metadata_sync = sync(pool_metadata)
symbol_address_sync = sync(symbol_address)
volume_sync = sync(volume)
reserves_sync = sync(reserves)
latest_reserves_sync = sync(latest_reserves)



"""
Tools for fetching pool state and metadata.

Currently supports stableswap pools, metapools, rebasing (RAI) metapools,
and 2-token cryptopools.
"""

__all__ = [
    "from_address",
    "get_data_cache",
    "get_metadata",
]

from curvesim.pool_data.metadata import PoolMetaData

from .cache import PoolDataCache
from .queries import from_address


def get_data_cache(address, chain="mainnet", days=60, end=None):
    """
    Fetch historical volume and redemption price data and return
    in a cache object.

    Deprecation warning: this will likely be removed in a future release.

    Parameters
    ----------
    address : str
        Pool address prefixed with “0x”.

    chain : str
        Chain/layer2 identifier, e.g. “mainnet”, “arbitrum”, “optimism".

    Returns
    -------
    :class:`PoolDataCache`

    """
    # TODO: validate function arguments
    metadata_dict = from_address(address, chain, end_ts=end)
    pool_data = PoolDataCache(metadata_dict, days=days, end=end)

    return pool_data


def get_metadata(
    address,
    chain="mainnet",
    env="prod",
    end_ts=None,
):
    """
    Pulls pool state and metadata from daily snapshot.

    Parameters
    ----------
    address : str
        Pool address prefixed with “0x”.

    chain : str
        Chain/layer2 identifier, e.g. “mainnet”, “arbitrum”, “optimism".

    Returns
    -------
    :class:`~curvesim.pool_data.metadata.PoolMetaDataInterface`

    """
    # TODO: validate function arguments
    metadata_dict = from_address(address, chain, env=env, end_ts=end_ts)
    metadata = PoolMetaData(metadata_dict)

    return metadata



"""
Network connector for on-chain data
"""

from asyncio import gather, sleep

from web3 import AsyncHTTPProvider, Web3
from web3.eth import AsyncEth

from curvesim.utils import get_env_var

from .http import HTTP
from .utils import sync

ETHERSCAN_URL = "https://api.etherscan.io/api"


def get_etherscan_api_key():
    """
    Get the Etherscan API key from the environment.
    Default to key provided by `curvesim` (not recommended).
    """
    default_key = "PT1D9IGAPPPRFMD312V9GARWW93BS9ZV6V"
    key = get_env_var("ETHERSCAN_API_KEY", default=default_key)
    return key


def get_alchemy_api_key():
    """
    Get the Alchemy API key from the environment.
    Default to key provided by `curvesim` (not recommended).
    """
    default_key = "WLcYLj9I1w7wEOgKmzidN1z62sbFILUz"
    key = get_env_var("ALCHEMY_API_KEY", default=default_key)
    return key


async def explorer(params):
    """
    Async function to retrieve data from the chain explorer (Etherscan).

    Parameters
    ----------
    params : dict
        Must include keys for module, action, and any required arguments for the query.

    Returns
    -------
    dict
        Query result

    """
    etherscan_api_key = get_etherscan_api_key()
    params.update({"apikey": etherscan_api_key})

    t_wait = 0.2
    while True:
        r = await HTTP.get(ETHERSCAN_URL, params=params)
        result = r["result"]

        if result.startswith("Max rate limit reached"):
            await sleep(t_wait)
            t_wait = round(t_wait * 1.5, 2)
        else:
            break

    return result


async def ABI(address):
    """
    Async function to retrieves ABI from the chain explorer (Etherscan).

    Parameters
    ----------
    address : str
        Address for the contract on Ethereum mainnet.

    Returns
    -------
    abi : str

    """
    p = {
        "module": "contract",
        "action": "getabi",
        "address": address,
    }

    abi = await explorer(p)

    return abi


_web3 = None


def _load_web3():
    alchemy_api_key = get_alchemy_api_key()
    global _web3  # pylint: disable=global-statement
    if not _web3:
        _web3 = Web3(
            AsyncHTTPProvider(
                f"https://eth-mainnet.g.alchemy.com/v2/{alchemy_api_key}",
                request_kwargs={"headers": {"Accept-Encoding": "gzip"}},
            ),
            modules={"eth": (AsyncEth,)},
            middlewares=[],
        )
    return _web3


async def contract(address, abi=None):
    """
    Creates an async Web3py contract object.

    Parameters
    ----------
    address : str
        Address for the contract on Ethereum mainnet.

    Returns
    -------
    contract : web3.contract.AsyncContract

    """
    abi = abi or await ABI(address)
    w3 = _load_web3()
    c = w3.eth.contract(address=address, abi=abi)
    return c


async def _underlying_coin_info(address):
    c = await contract(address)

    fns = ["upgradeToAndCall", "underlying", "token"]
    n_fns = len(fns) - 1

    for i, fn in enumerate(fns):
        if fn in dir(c.functions):
            break

        if i == n_fns:
            raise ValueError(f"Could not find underlying token for {address}")

    # Handle Aave proxy
    if fn == "upgradeToAndCall":
        abi = await ABI("0x1C050bCa8BAbe53Ef769d0d2e411f556e1a27E7B")
        c = await contract(address, abi)
        fn = "UNDERLYING_ASSET_ADDRESS"

    address, decimals = await gather(
        c.functions[fn]().call(), c.functions.decimals().call()
    )

    return address, decimals


async def underlying_coin_info(addresses):
    """
    Async function to get the underlying coin addresses for lending tokens
    (aTokens, cTokens, and yTokens).

    Parameters
    ----------
    addresses : iterable of str
        Addresses for the lending tokens on Ethereum mainnet.

    Returns
    -------
    addresses : list of str
        The addresses of the underlying tokens.

    """
    if isinstance(addresses, str):
        addrs, decimals = await _underlying_coin_info(addresses)

    else:
        tasks = []
        for address in addresses:
            tasks.append(_underlying_coin_info(address))

        response = await gather(*tasks)
        addrs, decimals = map(list, zip(*response))

    return addrs, decimals


# Sync
explorer_sync = sync(explorer)
ABI_sync = sync(ABI)
contract_sync = sync(contract)
underlying_coin_info_sync = sync(underlying_coin_info)



"""
Network connector for Coingecko.
"""
# pylint: disable=redefined-outer-name
import asyncio
from datetime import datetime, timedelta, timezone

import numpy as np
import pandas as pd

from curvesim.utils import get_pairs

from .http import HTTP
from .utils import sync

URL = "https://api.coingecko.com/api/v3/"

PLATFORMS = {
    "mainnet": "ethereum",
    "arbitrum": "arbitrum-one",
    "polygon": "polygon-pos",
    "optimism": "optimistic-ethereum",
    "xdai": "xdai",
    "fantom": "fantom",
    "avalanche": "avalanche",
    "matic:": "polygon-pos",
}


async def _get_prices(coin_id, vs_currency, start, end):
    url = URL + f"coins/{coin_id}/market_chart/range"
    p = {"vs_currency": vs_currency, "from": start, "to": end}

    r = await HTTP.get(url, params=p)

    return r


async def get_prices(coin_id, vs_currency, start, end):
    r = await _get_prices(coin_id, vs_currency, start, end)

    # Format data
    data = pd.DataFrame(r["prices"], columns=["timestamp", "prices"])
    data = data.merge(
        pd.DataFrame(r["total_volumes"], columns=["timestamp", "volumes"])
    )
    data["timestamp"] = pd.to_datetime(data["timestamp"], unit="ms", utc="True")
    data = data.set_index("timestamp")

    return data


async def _pool_prices(coins, vs_currency, days, end=None):
    if end is not None:
        # Times to reindex to: daily intervals
        # Coingecko only allows daily data when more than 90 days in the past
        # for the free REST endpoint
        t_end = datetime.fromtimestamp(end, tz=timezone.utc)
        t_start = t_end - timedelta(days=days + 1)
        t_samples = pd.date_range(start=t_start, end=t_end, freq="1D", tz=timezone.utc)
    else:
        # Times to reindex to: hourly intervals starting on half hour mark
        t_end = datetime.now(timezone.utc) - timedelta(days=1)
        t_end = t_end.replace(hour=23, minute=30, second=0, microsecond=0)
        t_start = t_end - timedelta(days=days + 1)
        t_samples = pd.date_range(start=t_start, end=t_end, freq="60T", tz=timezone.utc)
        end = t_end.timestamp()

    # Fetch data
    tasks = []
    for coin in coins:
        start = t_start.timestamp() - 86400 * 3
        tasks.append(get_prices(coin, vs_currency, start, end))

    data = await asyncio.gather(*tasks)

    # Format data
    qprices = []
    qvolumes = []
    for d in data:
        d.drop(d.tail(1).index, inplace=True)  # remove last row
        d = d.reindex(t_samples, method="ffill")
        qprices.append(d["prices"])
        qvolumes.append(d["volumes"])

    qprices = pd.concat(qprices, axis=1)
    qvolumes = pd.concat(qvolumes, axis=1)
    qvolumes = qvolumes / np.array(qprices)

    return qprices, qvolumes


def pool_prices(coins, vs_currency, days, chain="mainnet", end=None):
    """
    Pull price and volume data for given coins, quoted in given
    quote currency for given days.

    Parameters
    ----------
    coins: list of str
        List of coin addresses.
    vs_currency: str
        Symbol for quote currency.
    days: int
        Number of days to pull data for.

    Returns
    -------
    pair of pandas.Series
        prices Series and volumes Series
    """
    # Get data
    coins = coin_ids_from_addresses_sync(coins, chain)
    qprices, qvolumes = _pool_prices_sync(coins, vs_currency, days, end)

    # Compute prices by coin pairs
    combos = get_pairs(len(coins))
    prices = []
    volumes = []

    for pair in combos:
        base_price = qprices.iloc[:, pair[0]]
        base_volume = qvolumes.iloc[:, pair[0]]

        quote_price = qprices.iloc[:, pair[1]]
        quote_volume = qvolumes.iloc[:, pair[1]]

        # divide prices: (usd/base) / (usd/quote) = quote/base
        prices.append(base_price / quote_price)
        # sum volumes and convert to base: usd / (usd/base) = base
        volumes.append((base_volume + quote_volume) / base_price)

    prices = pd.concat(prices, axis=1)
    volumes = pd.concat(volumes, axis=1)

    return prices, volumes


async def _coin_id_from_address(address, chain):
    address = address.lower()
    chain = PLATFORMS[chain.lower()]
    url = URL + f"coins/{chain}/contract/{address}"

    r = await HTTP.get(url)

    coin_id = r["id"]

    return coin_id


async def coin_ids_from_addresses(addresses, chain):
    if isinstance(addresses, str):
        coin_ids = await _coin_id_from_address(addresses, chain)

    else:
        tasks = []
        for addr in addresses:
            tasks.append(_coin_id_from_address(addr, chain))

        coin_ids = await asyncio.gather(*tasks)

    return coin_ids


# Sync
_pool_prices_sync = sync(_pool_prices)
coin_ids_from_addresses_sync = sync(coin_ids_from_addresses)


if __name__ == "__main__":
    coin_addresses = [
        "0x6B175474E89094C44Da98b954EedeAC495271d0F",
        "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
        "0xdAC17F958D2ee523a2206206994597C13D831ec7",
    ]
    chain = "mainnet"
    print("Coin addresses:", coin_addresses)
    print("Chain", chain)

    vs_ccy = "USD"
    days = 1
    prices, volumes = pool_prices(coin_addresses, vs_ccy, days, chain)
    print(prices.head())
    print(volumes.head())



from curvesim.logging import get_logger
from curvesim.templates.trader import Trade, Trader

from ..common import get_arb_trades

logger = get_logger(__name__)


class SimpleArbitrageur(Trader):
    """
    Computes, executes, and reports out arbitrage trades.
    """

    # pylint: disable-next=arguments-differ,too-many-locals
    def compute_trades(self, prices):
        """
        Compute trades to arbitrage the pool, as follows:
            1. For each coin pair i and j, calculate size of coin i
               needed to move price of coin i w.r.t. to j to the
               target price.
            2. Calculate the profit from each such swap.
            3. Take the swap that gives the largest profit.

        Parameters
        ----------
        prices : pandas.Series
            Current market prices from the price_sampler.

        Returns
        -------
        trades : list of :class:`Trade` objects
            List of trades to perform.

        additional_data: dict
            Dict of additional data to be passed to the state log as part of trade_data.
        """
        pool = self.pool
        trades = get_arb_trades(pool, prices)

        max_profit = 0
        best_trade = None
        price_error = None
        for t in trades:
            size, coins, price_target = t
            i, j = coins
            min_trade_size = pool.get_min_trade_size(i)
            if size <= min_trade_size:
                continue
            with pool.use_snapshot_context():
                out_amount, _ = pool.trade(i, j, size)
                # assume we transacted at "infinite" depth at target price
                # on the other exchange to obtain our in-token
                profit = out_amount - size * price_target
                if profit > max_profit:
                    max_profit = profit
                    best_trade = Trade(i, j, size)
                    price_error = pool.price(i, j) - price_target

        if not best_trade:
            return [], {"price_errors": []}

        return [best_trade], {"price_errors": [price_error]}



"""
Pure calculations with a common interface to wrap around
differing calculations for 2 or coins.
"""
__all__ = [
    "geometric_mean",
    "newton_D",
    "get_y",
    "get_alpha",
    "get_p",
    "halfpow",
]
from math import ceil
from typing import List

from gmpy2 import mpz

from curvesim.exceptions import CalculationError, CurvesimValueError
from curvesim.logging import get_logger

from . import factory_2_coin, tricrypto_ng
from .tricrypto_ng import get_p

EXP_PRECISION = 10**10

logger = get_logger(__name__)


def geometric_mean(unsorted_x: List[int]) -> int:
    """
    (x[0] * x[1] * ...) ** (1/N)
    """
    n_coins = len(unsorted_x)
    if n_coins == 2:
        mean = factory_2_coin.geometric_mean(unsorted_x, True)
    elif n_coins == 3:
        mean = tricrypto_ng.geometric_mean(unsorted_x)
    else:
        raise CurvesimValueError("More than 3 coins is not supported.")

    return mean


def newton_D(A: int, gamma: int, xp: List[int], K0_prev: int = 0) -> int:
    """
    Compute D using using specific approaches depending on
    the number of coins.

    For 3 coins, we actually use Halley's method and allow a
    starting value.
    """
    n_coins = len(xp)
    if n_coins == 2:
        D = factory_2_coin.newton_D(A, gamma, xp)
    elif n_coins == 3:
        D = tricrypto_ng.newton_D(A, gamma, xp, K0_prev)
    else:
        raise CurvesimValueError("More than 3 coins is not supported.")

    return D


def get_y(A: int, gamma: int, xp: List[int], D: int, j: int) -> List[int]:
    n_coins = len(xp)
    if n_coins == 2:
        y_out = [factory_2_coin.newton_y(A, gamma, xp, D, j), 0]
    elif n_coins == 3:
        y_out = tricrypto_ng.get_y(A, gamma, xp, D, j)
    else:
        raise CurvesimValueError("More than 3 coins is not supported.")

    return y_out


def get_alpha(ma_half_time, block_timestamp, last_prices_timestamp, n_coins):
    if n_coins == 2:
        alpha: int = halfpow(
            (block_timestamp - last_prices_timestamp) * 10**18 // ma_half_time
        )
    elif n_coins == 3:
        # tricrypto-ng stores the ma half-time divided by ln(2), so we have to
        # take the real half-time and divide by ln(2) to use in the alpha calc.
        #
        # Note ln(2) = 0.693147... but the approx actually used is 694 / 1000.
        #
        # CAUTION: neeed to be wary of off-by-one errors from integer division.
        ma_half_time = ceil(ma_half_time * 1000 / 694)
        alpha: int = tricrypto_ng.wad_exp(
            -1 * ((block_timestamp - last_prices_timestamp) * 10**18 // ma_half_time)
        )
    else:
        raise CurvesimValueError("More than 3 coins is not supported.")

    return alpha


def halfpow(power: int) -> int:
    """
    1e18 * 0.5 ** (power/1e18)

    Inspired by:
    https://github.com/balancer-labs/balancer-core/blob/master/contracts/BNum.sol#L128
    """
    intpow: int = power // 10**18
    otherpow: int = power - intpow * 10**18
    if intpow > 59:
        return 0
    result: int = 10**18 // (2**intpow)
    if otherpow == 0:
        return result

    term: int = mpz(10**18)
    x: int = 5 * 10**17
    S: int = 10**18
    neg: bool = False

    for i in range(1, 256):
        K: int = i * 10**18
        c: int = K - 10**18
        if otherpow > c:
            c = otherpow - c
            neg = not neg
        else:
            c -= otherpow
        term = term * (c * x // 10**18) // K
        if neg:
            S -= term
        else:
            S += term
        if term < EXP_PRECISION:
            return int(result * S // 10**18)

    raise CalculationError("Did not converge")



from copy import deepcopy

import pytest
from pandas import DataFrame, Series

from curvesim.exceptions import MetricError
from curvesim.metrics import init_metrics, metrics
from curvesim.metrics.base import Metric, PoolMetric, PricingMetric, get_coin_pairs

all_metrics = [getattr(metrics, m) for m in metrics.__all__]


def test_init_metrics(
    sim_curve_pool, sim_curve_crypto_pool, sim_curve_meta_pool, sim_curve_rai_pool
):
    """Test that all metrics successfully initialize with correct attributes/properties."""

    pools = [
        sim_curve_pool,
        sim_curve_crypto_pool,
        sim_curve_meta_pool,
        sim_curve_rai_pool,
    ]

    coin_names = ["COIN0", "COIN1"]
    coin_addresses = ["0x0", "0x1"]
    prices = {("COIN0", "COIN1"): 10}

    metadata = {
        "coins": {
            "names": coin_names,
            "addresses": coin_addresses,
        },
        "chain": "mainnet",
    }

    basepool_metadata = {
        "coins": {
            "names": [coin_names[-1]],
            "addresses": [coin_addresses[-1]],
        },
        "chain": "mainnet",
    }

    for pool in pools:
        # Set metadata
        pool.metadata = metadata
        if hasattr(pool, "basepool"):
            pool.basepool.metadata = basepool_metadata

        # Loop through metric tests
        metrics = init_metrics(all_metrics, pool=pool)
        for metric in metrics:
            _test_metric_class_init(metric)

            if isinstance(metric, PoolMetric):
                _test_pool_metric_class_init(metric, pool)

            if isinstance(metric, PricingMetric):
                _test_pricing_metric_class_init(metric, coin_names, prices)


def _test_metric_class_init(metric):
    """Test correct mapping of properties for Metric (sub)class."""
    assert metric.metric_function == metric.config["functions"]["metrics"]
    assert metric.plot_config == metric.config.get("plot", None)
    assert metric.summary_functions == metric.config["functions"].get("summary", None)


def _test_pool_metric_class_init(metric, pool):
    """Test properties and methods specific to PoolMetric (sub)class."""

    # Test properties
    assert metric._pool == pool
    assert metric.config == metric.pool_config[type(pool)]

    # Test set_pool
    pool_copy = deepcopy(pool)
    metric.set_pool(pool_copy)
    assert metric._pool == pool_copy
    assert metric._pool != pool

    # Test set_pool_state
    metric.set_pool_state({"A": 0})
    assert metric._pool.A == 0
    if hasattr(pool, "basepool"):
        metric.set_pool_state({"A_base": 0})
        assert metric._pool.basepool.A == 0


def _test_pricing_metric_class_init(metric, coin_names, prices):
    """Test attributes and methods specific to PricingMetric (sub)class."""

    # Test PricingMetric attributes
    assert metric.coin_names == coin_names
    assert metric.numeraire == coin_names[0]

    # Test get_market_price
    assert metric.get_market_price("COIN0", "COIN1", prices) == 10
    assert metric.get_market_price("COIN1", "COIN0", prices) == 1 / 10


def test_get_poin_pairs():
    """Test that get_coin_pairs returns same results regardless of input format."""
    price_dict = {("COIN0", "COIN1"): 10}
    price_df = DataFrame.from_records([price_dict])
    price_df2 = DataFrame.from_records([price_dict, price_dict])
    price_series = Series(price_dict)

    formats = [price_dict, price_df, price_df2, price_series]

    for i in range(len(formats) - 1):
        format1 = get_coin_pairs(formats[i])
        format2 = get_coin_pairs(formats[i + 1])
        assert format1 == format2
        assert type(format1) == type(format2)


def test_get_coin_pairs_wrong_type():
    """Test that unsupported input type raises error."""
    inputs = ["123", 123, [1, 2, 3], (1, 2, 3), {1, 2, 3}]

    for inp in inputs:
        with pytest.raises(MetricError):
            get_coin_pairs(inp)


def test_missing_metric_function_exception():
    """Test that metric with no metric_function raises error."""

    class WorkingDummyMetric(Metric):
        @property
        def config(self):
            return {"functions": {"metrics": "working_function"}}

    class BrokenDummyMetric(Metric):
        @property
        def config(self):
            return {"functions": {}}

    metric1 = WorkingDummyMetric()
    metric2 = BrokenDummyMetric()

    assert metric1.metric_function == "working_function"
    with pytest.raises(MetricError):
        metric2.metric_function



from abc import ABC, abstractmethod


class PoolMetaDataInterface(ABC):
    """
    Interface for container holding a pool's metadata.

    Besides easy to access attributes, it provides `init_kwargs`, which
    are keyword arguments to use in a pool's `__init__`.  This is useful
    for factory functions.
    """

    @abstractmethod
    def init_kwargs(self, balanced=True, balanced_base=True, normalize=True):
        """
        Returns a dictionary of kwargs used to initialize the pool.

        Parameters
        ----------
        balanced: bool
            Will derive balances corresponding to the pool's total value so
            that it is balanced.
        balanced_base: bool
            For metapools only.

            Will derive balances corresponding to the basepool's total value so
            that it is balanced.
        normalize: bool
            Will put the balances in units of D, e.g. 18 decimals.

        Returns
        -------
        dict
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def address(self):
        """
        Returns the pool address.

        Returns
        -------
        str
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def chain(self):
        """
        Returns the chain/layer 2 identifier.

        Returns
        -------
        str
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def pool_type(self):
        """
        Returns the pool type.

        Returns
        -------
        Pool

        """
        raise NotImplementedError

    @property
    @abstractmethod
    def sim_pool_type(self):
        """
        Returns the pool type.

        Returns
        -------
        SimPool

        """
        raise NotImplementedError

    @property
    @abstractmethod
    def coins(self):
        """
        Returns coin addresses for the pool's holdings.

        For pools that are not on Ethereum mainnet, the address
        for the corresponding mainnet token is returned.

        For lending tokens (e.g., aTokens or cTokens), the
        address for the underlying token is returned.

        Returns
        -------
        list of strings
            coin addresses

        """
        raise NotImplementedError

    @property
    @abstractmethod
    def coin_names(self):
        """
        Returns coin names for the pool's holdings.

        For pools that are not on Ethereum mainnet, the name
        of the corresponding mainnet token is returned.

        For lending tokens (e.g., aTokens or cTokens), the
        name of the underlying token is returned.

        Returns
        -------
        list of strings
            coin names

        """
        raise NotImplementedError

    @property
    @abstractmethod
    def n(self):
        """
        Returns the number of token-types (e.g., DAI, USDC, USDT) in a pool.

        Returns
        -------
        int or list of ints
            Number of token-types.

            For metapools, a list [n_metapool, n_basepool] is returned.

            N_metapool includes the basepool LP token.

        """
        raise NotImplementedError


class PoolMetaDataBase(PoolMetaDataInterface, ABC):
    """
    Base class for pool metadata containers.

    Provides a very simple and generic implementation that is used
    by simulation pipelines.
    """

    def __init__(self, metadata_dict, pool_type, sim_pool_type):
        self._dict = metadata_dict
        self._pool_type = pool_type
        self._sim_pool_type = sim_pool_type

    @property
    def address(self):
        return self._dict["address"]

    @property
    def chain(self):
        return self._dict["chain"]

    @property
    def pool_type(self):
        return self._pool_type

    @property
    def sim_pool_type(self):
        return self._sim_pool_type
